# EVALUATOR - PROGRESS MANAGER
START_CONFIG = 'START_CONFIG'
END_CONFIG = 'END_CONFIG'
END_FINAL_RUN = 'END_FINAL_RUN'
OUTER_FOLD = 'outer_fold'
INNER_FOLD = 'inner_fold'
CONFIG_ID = 'config_id'
RUN_ID = 'run_id'
ELAPSED = 'elapsed'
EXP_NAME = 'exp_name'
SEED = 'seed'
GRID_SEARCH = 'grid'
RANDOM_SEARCH = 'random'
NUM_SAMPLES = 'num_samples'
SAMPLE_METHOD = 'sample_method'
DATASET = 'dataset'
CONFIG = 'config'
MODEL_ASSESSMENT = 'MODEL_ASSESSMENT'
FOLDS = 'folds'
TR_LOSS = 'training_loss'
VL_LOSS = 'validation_loss'
TR_SCORE = 'training_score'
VL_SCORE = 'validation_score'
BEST_CONFIG = 'best_config'
OUTER_TRAIN = 'outer_train'
OUTER_VALIDATION = 'outer_validation'
OUTER_TEST = 'outer_test'

# SPLITTER
DATA_SPLITTER_BASE_PATH = 'pydgn.data.splitter.'

# ENGINE
DEFAULT_ENGINE_CALLBACK = 'pydgn.training.callback.engine_callback.EngineCallback'

# STATE
TRAINING = 'training'
EVALUATION = 'evaluation'
VALIDATION = 'validation'
TEST = 'test'
MAIN_LOSS = 'main_loss'
MAIN_SCORE = 'main_score'
LOSSES = 'losses'
LOSS = 'loss'
SCORES = 'scores'
SCORE = 'score'
MODEL_STATE = 'model_state'
OPTIMIZER_STATE = 'optimizer_state'
SCHEDULER_STATE = 'scheduler_state'
BEST_EPOCH = 'best_epoch'
BEST_EPOCH_RESULTS = 'best_epoch_results'
EPOCH = 'epoch'
STOP_TRAINING = 'stop_training'
LAST_CHECKPOINT_FILENAME = 'last_checkpoint.pth'
BEST_CHECKPOINT_FILENAME = 'best_checkpoint.pth'
BATCH_LOSS_EXTRA = 'batch_loss_extra'

# STRING FORMATTING
TENSORBOARD = 'tensorboard'
EMB_TUPLE_SUBSTR = '_embeddings_tuple'
ATOMIC_SAVE_EXTENSION = '.part'
CLASS_NAME = 'class_name'
ARGS = 'args'
MODEL_STATE_DICT = 'model_state_dict'
OPTIMIZER_STATE_DICT = 'optimizer_state_dict'
AVG = 'avg'
STD = 'std'

# SET UP
CUDA = 'cuda'
PYDGN_RAY_NUM_GPUS_PER_TASK = 'PYDGN_RAY_NUM_GPUS_PER_TASK'
MIN = 'min'
MAX = 'max'

# CLI
CONFIG_FILE_CLI_ARGUMENT = '--config-file'
CONFIG_FILE = 'config_file'
DEBUG_CLI_ARGUMENT = '--debug'

DATA_ROOT = 'data_root'
DATA_SPLITS_FILE = 'data_splits_file'
DATASET_CLASS = 'dataset_class'
DATASET_GETTER = 'dataset_getter'
evaluate_every = 'evaluate_every'
EVALUATE_EVERY = 'evaluate_every'
DATA_LOADER = 'data_loader'
DATA_LOADER_ARGS = 'data_loader_args'
DATASET_NAME = 'dataset_name'
DEBUG = 'debug'
DEVICE = 'device'
EXPERIMENT = 'experiment'
FINAL_TRAINING_RUNS = 'final_training_runs'
GPUS_PER_TASK = 'gpus_per_task'
HIGHER_RESULTS_ARE_BETTER = 'higher_results_are_better'
MAX_CPUS = 'max_cpus'
MAX_GPUS = 'max_gpus'
MODEL = 'model'
RESULT_FOLDER = 'result_folder'


from pydgn.experiment.util import s2c
from pydgn.training.event.handler import EventHandler
from pydgn.training.event.state import State


class GradientClipper(EventHandler):
    r"""
    GradientClipper is the main event handler for gradient clippers. Just pass a PyTorch scheduler together with its
    arguments in the configuration file.

    Args:
        gradient_clipper_class_name (str): the dotted path to the gradient clipper class name
        kwargs (dict): additional arguments
    """
    def __init__(self, gradient_clipper_class_name: str, **kwargs: dict):
        self.gradient_clipper = s2c(gradient_clipper_class_name)(**kwargs)

    def on_backward(self, state: State):
        self.gradient_clipper.clip_gradients(state.model.parameters())



import math
import random
import warnings
from typing import Union, Callable

import numpy as np
import torch
import torch_geometric.loader
from torch.utils.data import Subset
from torch_geometric.data import Data, Batch
from torch_geometric.loader.dataloader import Collater

import pydgn.data.dataset
from pydgn.data.dataset import DatasetInterface
from pydgn.data.sampler import RandomSampler
from pydgn.data.splitter import Splitter, LinkPredictionSingleGraphSplitter
from pydgn.data.dataset import TemporalDatasetInterface
from pydgn.data.util import load_dataset


def seed_worker(exp_seed, worker_id):
    r"""
    Used to set a different, but reproducible, seed for all data-retriever workers. Without this,
    all workers will retrieve the data in the same order.

    Args:
        exp_seed (int): base seed to be used for reproducibility
        worker_id (int): id number of the worker
    """
    np.random.seed(exp_seed + worker_id)
    random.seed(exp_seed + worker_id)


class DataProvider:
    r"""
    A DataProvider object retrieves the correct data according to the external and internal data splits.
    It can be additionally used to augment the data, or to create a specific type of data loader. The base class
    does nothing special, but here is where the i-th element of a dataset could be pre-processed before constructing
    the mini-batches.

    Args:
        data_root (str): the path of the root folder in which data is stored
        splits_filepath (str): the filepath of the splits. with additional metadata
        dataset_class (Callable[...,:class:`pydgn.data.dataset.DatasetInterface`]): the class of the dataset
        data_loader_class (Union[Callable[...,:class:`torch.utils.data.DataLoader`], Callable[...,:class:`torch_geometric.loader.DataLoader`]]): the class of the data loader to use
        data_loader_args (dict): the arguments of the data loader
        dataset_name (str): the name of the dataset
        outer_folds (int): the number of outer folds for risk assessment. 1 means hold-out, >1 means k-fold
        inner_folds (int): the number of outer folds for model selection. 1 means hold-out, >1 means k-fold

    """
    def __init__(self,
                 data_root: str,
                 splits_filepath: str,
                 dataset_class: Callable[...,pydgn.data.dataset.DatasetInterface],
                 dataset_name: str,
                 data_loader_class: Union[Callable[...,torch.utils.data.DataLoader],
                                          Callable[...,torch_geometric.loader.DataLoader]],
                 data_loader_args: dict,
                 outer_folds: int,
                 inner_folds: int):

        self.exp_seed = None

        self.data_root = data_root
        self.dataset_class = dataset_class
        self.dataset_name = dataset_name

        self.data_loader_class = data_loader_class
        self.data_loader_args = data_loader_args

        self.outer_folds = outer_folds
        self.inner_folds = inner_folds

        self.outer_k = None
        self.inner_k = None

        self.splits_filepath = splits_filepath
        self.splitter = None
        self.dataset = None  # use this to avoid instantiating multiple versions of the same dataset when no run-time specific arguments are needed

        self.dim_node_features = None
        self.dim_edge_features = None
        self.dim_target = None


    def set_exp_seed(self, seed: int):
        r"""
        Sets the experiment seed to give to the DataLoader. Helps with reproducibility.

        Args:
            seed (int): id of the seed
        """
        self.exp_seed = seed

    def set_outer_k(self, k: int):
        r"""
        Sets the parameter k of the `risk assessment` procedure. Called by the evaluation modules to load the correct
        subset of the data.

        Args:
            k (int): the id of the fold, ranging from 0 to K-1.
        """
        self.outer_k = k

    def set_inner_k(self, k):
        r"""
        Sets the parameter k of the `model selection` procedure. Called by the evaluation modules to load the correct
        subset of the data.

        Args:
            k (int): the id of the fold, ranging from 0 to K-1.
        """
        self.inner_k = k

    def _get_splitter(self) -> Splitter:
        """
        Instantiates the splitter with the parameters stored in the file ``self.splits_filepath``

        Returns:
            a :class:`~pydgn.data.splitter.Splitter` object
        """
        if self.splitter is None:
            self.splitter = Splitter.load(self.splits_filepath)
        return self.splitter

    def _get_dataset(self, **kwargs: dict) -> DatasetInterface:
        """
        Instantiates the dataset. Relies on the parameters stored in the ``dataset_kwargs.pt`` file.

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed to the dataset. Not used in the base version

        Returns:
            a :class:`~pydgn.data.dataset.DatasetInterface` object
        """
        assert 'shuffle' not in kwargs, "Your implementation of _get_loader should remove `shuffle` from kwargs before calling _get_dataset"
        assert 'batch_size' not in kwargs, "Your implementation of _get_loader should remove `batch_size` from kwargs before calling _get_dataset"

        if kwargs is not None and len(kwargs) != 0:
            # we probably need to pass run-time specific parameters, so load the dataset in memory again
            # an example is the subset of urls in Iterable style datasets
            dataset = load_dataset(self.data_root, self.dataset_name, self.dataset_class, **kwargs)
        else:
            if self.dataset is None:
                dataset = load_dataset(self.data_root, self.dataset_name, self.dataset_class)
                self.dataset = dataset
            else:
                dataset = self.dataset

        self.dim_node_features = dataset.dim_node_features
        self.dim_edge_features = dataset.dim_edge_features
        self.dim_target = dataset.dim_target

        return dataset

    def _get_loader(self, indices: list, **kwargs: dict) -> Union[torch.utils.data.DataLoader,
                                                                  torch_geometric.loader.DataLoader]:
        r"""
        Instantiates the data loader.

        Args:
            indices (sequence): Indices in the whole set selected for subset
            kwargs (dict): a dictionary of additional arguments to be passed to the dataset being loaded.
                           Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`, :class:`torch_geometric.loader.DataLoader`] object
        """
        shuffle = kwargs.pop("shuffle", False)
        batch_size = kwargs.pop("batch_size", 1)

        dataset: DatasetInterface = self._get_dataset(**kwargs)
        dataset = Subset(dataset, indices)

        assert self.exp_seed is not None, "DataLoader's seed has not been specified! Is this a bug?"
        kwargs['worker_init_fn'] = lambda worker_id: seed_worker(worker_id, self.exp_seed)
        kwargs.update(self.data_loader_args)

        if shuffle is True:
            sampler = RandomSampler(dataset)
            dataloader = self.data_loader_class(dataset, sampler=sampler, batch_size=batch_size, **kwargs)
        else:
            dataloader = self.data_loader_class(dataset, shuffle=False, batch_size=batch_size, **kwargs)

        return dataloader

    def get_inner_train(self, **kwargs: dict) -> Union[torch.utils.data.DataLoader,
                                                 torch_geometric.loader.DataLoader]:
        r"""
        Returns the training set for model selection associated with specific outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed to the dataset being loaded.
                           Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`, :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None and self.inner_k is not None
        splitter = self._get_splitter()
        indices = splitter.inner_folds[self.outer_k][self.inner_k].train_idxs
        return self._get_loader(indices, **kwargs)

    def get_inner_val(self, **kwargs: dict) -> Union[torch.utils.data.DataLoader,
                                                torch_geometric.loader.DataLoader]:
        r"""
        Returns the validation set for model selection associated with specific outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed to the dataset being loaded.
                           Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`, :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None and self.inner_k is not None
        splitter = self._get_splitter()
        indices = splitter.inner_folds[self.outer_k][self.inner_k].val_idxs
        return self._get_loader(indices, **kwargs)

    def get_outer_train(self, **kwargs: dict) -> Union[torch.utils.data.DataLoader,
                                                                  torch_geometric.loader.DataLoader]:
        r"""
        Returns the training set for risk assessment associated with specific outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed to the dataset being loaded.
                           Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`, :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()

        train_indices = splitter.outer_folds[self.outer_k].train_idxs
        return self._get_loader(train_indices, **kwargs)

    def get_outer_val(self, **kwargs: dict) -> Union[torch.utils.data.DataLoader,
                                                                  torch_geometric.loader.DataLoader]:
        r"""
        Returns the validation set for risk assessment associated with specific outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed to the dataset being loaded.
                           Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`, :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()
        val_indices = splitter.outer_folds[self.outer_k].val_idxs
        return self._get_loader(val_indices, **kwargs)

    def get_outer_test(self, **kwargs: dict) -> Union[torch.utils.data.DataLoader,
                                                torch_geometric.loader.DataLoader]:
        r"""
        Returns the test set for risk assessment associated with specific outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed to the dataset being loaded.
                           Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`, :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()
        indices = splitter.outer_folds[self.outer_k].test_idxs
        return self._get_loader(indices, **kwargs)

    def get_dim_node_features(self) -> int:
        r"""
        Returns the number of node features of the dataset

        Returns:
            the value of the property ``dim_node_features`` in the dataset

        """
        if self.dim_node_features is None:
            raise Exception("You should first initialize the dataset by creating a data loader!")
        return self.dim_node_features

    def get_dim_edge_features(self) -> int:
        r"""
        Returns the number of node features of the dataset

        Returns:
            the value of the property ``dim_edge_features`` in the dataset

        """
        if self.dim_edge_features is None:
            raise Exception("You should first initialize the dataset by creating a data loader!")
        return self.dim_edge_features

    def get_dim_target(self) -> int:
        r"""
        Returns the dimension of the target for the task

        Returns:
            the value of the property ``dim_target`` in the dataset

        """
        if self.dim_target is None:
            raise Exception("You should first initialize the dataset by creating a data loader!")
        return self.dim_target


class IterableDataProvider(DataProvider):
    r"""
    A DataProvider object that allows to fetch data from an Iterable-style Dataset (see :class:`pydgn.data.dataset.IterableDatasetInterface`).
    """

    def _get_loader(self, indices: list, **kwargs: dict) -> Union[torch.utils.data.DataLoader,
                                                                  torch_geometric.loader.DataLoader]:

        shuffle = kwargs.pop("shuffle", False)
        batch_size = kwargs.pop("batch_size", 1)

        # we will overwrite the dataset each time the loader is called
        dataset = self._get_dataset(**{'url_indices': indices})

        if shuffle:
            dataset.shuffle_urls(True)
            dataset.shuffle_urls_elements(True)

        assert self.exp_seed is not None, "DataLoader's seed has not been specified! Is this a bug?"

        # Define a worker_init_fn that configures each dataset copy differently
        # this is called only when num_workers is set to a value > 0
        def worker_init_fn(worker_id):
            worker_info = torch.utils.data.get_worker_info()
            num_workers = worker_info.num_workers
            assert num_workers > 0

            # Set the random seed
            seed_worker(worker_id, self.exp_seed)

            # Get the dataset and overall length
            dataset = worker_info.dataset  # the dataset copy in this worker process
            dataset_length = len(dataset)  # dynamic, already refers to the subset of urls!

            per_worker = int(math.ceil((dataset_length) / float(worker_info.num_workers)))

            start = worker_id * per_worker
            end = worker_id * per_worker + per_worker

            # configure the dataset to only process the split workload
            dataset.splice(start, end)

        kwargs.update(self.data_loader_args)

        dataloader = self.data_loader_class(dataset, sampler=None, collate_fn=Collater(None, None),
                                            worker_init_fn=worker_init_fn, batch_size=batch_size, **kwargs)
        return dataloader


class SingleGraphSequenceDataProvider(DataProvider):
    """
    This class is responsible for building the dynamic dataset at runtime.
    """

    @classmethod
    def collate_fn(cls, samples_list):
        return [Batch.from_data_list([d]) for d in samples_list]

    def _get_loader(self, indices: list, **kwargs: dict) -> Union[torch.utils.data.DataLoader,
                                                                  torch_geometric.loader.DataLoader]:
        shuffle = kwargs.pop("shuffle", False)
        batch_size = kwargs.pop("batch_size", 1)

        dataset: TemporalDatasetInterface = self._get_dataset(**kwargs)
        dataset = Subset(dataset, indices)

        assert self.exp_seed is not None, 'DataLoader seed has not been specified! Is this a bug?'
        kwargs['worker_init_fn'] = lambda worker_id: seed_worker(worker_id, self.exp_seed)

        # Using Pytorch default DataLoader instead of PyG, to return list of graphs
        assert self.data_loader_class == torch.utils.data.DataLoader, "SingleGraphSequenceDataProvider accepts a torch DataLoader only!"
        dataloader = self.data_loader_class(dataset, collate_fn=SingleGraphSequenceDataProvider.collate_fn,
                                            shuffle=shuffle, batch_size=batch_size, **kwargs)

        kwargs.update(self.data_loader_args)
        return dataloader


class LinkPredictionSingleGraphDataProvider(DataProvider):
    r"""
    An extension of the DataProvider class to deal with link prediction on a single graph.
    Designed to work with :class:`~pydgn.data.splitter.LinkPredictionSingleGraphSplitter`.
    We also assume the single-graph dataset can fit in memory
    **WARNING**: this class **modifies** the dataset by creating copies. It may not work if a "shared dataset" feature
    is added to PyDGN.
    """

    def _get_dataset(self, **kwargs):
        assert 'shuffle' not in kwargs, "Your implementation of _get_loader should remove `shuffle` from kwargs before calling _get_dataset"
        assert 'batch_size' not in kwargs, "Your implementation of _get_loader should remove `batch_size` from kwargs before calling _get_dataset"

        # Since we modify the dataset, we need different istances of the same graph
        return load_dataset(self.data_root, self.dataset_name, self.dataset_class, **kwargs)

    def _get_splitter(self):
        super()._get_splitter()  # loads splitter into self.splitter
        assert isinstance(self.splitter,
                          LinkPredictionSingleGraphSplitter), "This class only work with a LinkPredictionSingleGraphSplitter splitter."
        return self.splitter

    def _get_loader(self, indices, **kwargs):
        # We may want to shuffle the edges of our single graph and take edge batches
        # NOTE: EVAL edges can be TRAINING/VAL/TEST. It is on "eval" edges
        # that we compute the loss (and eventually do training)
        # changing names may be good
        shuffle = kwargs.pop("shuffle", False)
        batch_size = kwargs.pop("batch_size", 1)

        dataset = self._get_dataset(**kwargs)  # Custom implementation, we need a copy of the dataset every time
        assert len(dataset) == 1, f"Provider accepts a single-graph dataset only, but I see {len(dataset)} graphs"
        y = dataset.data.y

        # Use indices to change edge_index and provide an y
        train, eval = indices
        pos_train_edges, train_attr, neg_train_edges = train
        pos_train_edges = torch.tensor(pos_train_edges, dtype=torch.long)
        train_attr = torch.tensor(train_attr) if train_attr is not None else None
        neg_train_edges = torch.tensor(neg_train_edges, dtype=torch.long)

        pos_eval_edges, eval_attr, neg_eval_edges = eval
        pos_eval_edges = torch.tensor(pos_eval_edges, dtype=torch.long)
        eval_attr = torch.tensor(eval_attr) if eval_attr is not None else None
        neg_eval_edges = torch.tensor(neg_eval_edges, dtype=torch.long)

        # The code below works because we are working with a single graph!
        dataset.data.edge_index = pos_train_edges
        dataset.data.edge_attr = train_attr
        # Eval can be training/val/test
        dataset.data.y = (y, pos_eval_edges, neg_eval_edges)

        if shuffle is True:
            pos_edge_indices = torch.randperm(pos_eval_edges.shape[1])
            neg_edge_indices = torch.randperm(neg_eval_edges.shape[1])
        else:
            pos_edge_indices = torch.arange(pos_eval_edges.shape[1])
            neg_edge_indices = torch.arange(neg_eval_edges.shape[1])

        # create batch of edges here, leaving other fields as is
        batched_edge_dataset = []

        permuted_pos_eval_edges = pos_eval_edges[:, pos_edge_indices]
        permuted_neg_eval_edges = neg_eval_edges[:, neg_edge_indices]
        permuted_edge_attr = eval_attr[pos_edge_indices] if eval_attr is not None else None
        done = False
        batch_start = 0
        while not done:
            if batch_size == 0:
                # Full batch
                pos_batch_end = -1
                neg_batch_end = -1
            else:
                # Create subgraph
                pos_batch_end = batch_start + batch_size if (batch_start + batch_size) < permuted_pos_eval_edges.shape[
                    1] else -1
                neg_batch_end = batch_start + batch_size if (batch_start + batch_size) < permuted_neg_eval_edges.shape[
                    1] else -1

            batch_pos_edge_indices = permuted_pos_eval_edges[:, batch_start:pos_batch_end]
            batch_neg_edge_indices = permuted_neg_eval_edges[:, batch_start:neg_batch_end]
            batch_edge_attr = permuted_edge_attr[batch_start:pos_batch_end] if eval_attr is not None else None
            batch_start += batch_size

            if pos_batch_end == -1 or neg_batch_end == -1:  # to ensure balancing
                done = True

            # create data object and append to list
            subgraph = Data(x=dataset.data.x, edge_index=dataset.data.edge_index,
                            edge_attr=dataset.data.edge_attr,
                            y=(y, batch_pos_edge_indices, batch_neg_edge_indices))
            batched_edge_dataset.append(subgraph)

        assert self.exp_seed is not None, 'DataLoader seed has not been specified! Is this a bug?'
        kwargs['worker_init_fn'] = lambda worker_id: seed_worker(worker_id, self.exp_seed)
        kwargs.update(self.data_loader_args)

        # Single graph dataset, shuffle does not make sense (unless we know how to do mini-batch training with nodes)
        dataloader = self.data_loader_class(batched_edge_dataset, batch_size=1, shuffle=False,
                                            **kwargs)

        return dataloader

    def get_inner_train(self, **kwargs):
        assert self.outer_k is not None and self.inner_k is not None
        splitter = self._get_splitter()
        train_indices = splitter.inner_folds[self.outer_k][self.inner_k].train_idxs
        eval_indices = train_indices
        indices = train_indices, eval_indices
        return self._get_loader(indices, **kwargs)

    def get_inner_val(self, **kwargs):
        assert self.outer_k is not None and self.inner_k is not None
        splitter = self._get_splitter()
        train_indices = splitter.inner_folds[self.outer_k][self.inner_k].train_idxs
        eval_indices = splitter.inner_folds[self.outer_k][self.inner_k].val_idxs
        indices = train_indices, eval_indices
        return self._get_loader(indices, **kwargs)

    def get_outer_train(self, **kwargs):
        assert self.outer_k is not None
        splitter = self._get_splitter()

        train_indices = splitter.outer_folds[self.outer_k].train_idxs
        eval_indices = train_indices
        indices = train_indices, eval_indices
        return self._get_loader(indices, **kwargs)

    def get_outer_val(self, **kwargs):
        assert self.outer_k is not None
        splitter = self._get_splitter()

        train_indices = splitter.outer_folds[self.outer_k].train_idxs
        eval_indices = splitter.outer_folds[self.outer_k].val_idxs

        indices = train_indices, eval_indices
        return self._get_loader(indices, **kwargs)

    def get_outer_test(self, **kwargs):
        assert self.outer_k is not None
        splitter = self._get_splitter()
        train_indices = splitter.outer_folds[self.outer_k].train_idxs
        eval_indices = splitter.outer_folds[self.outer_k].test_idxs
        indices = train_indices, eval_indices
        return self._get_loader(indices, **kwargs)


# TBD: taken from old branch
# class MultipleGraphSequenceDataProvider(DataProvider):
#     """
#     IMPORTANT: This provider assumes all graph sequences have the same length!
#     """
#
#     @classmethod
#     def collate_fn(cls, samples_list):
#         # Decompose by timestep and graph sequence, then batch graphs belonging
#         # to the same time step
#
#         num_sequences = len(samples_list)
#         assert num_sequences >= 1
#
#         # Search for the maximum sequence length
#         num_timesteps = 0
#         for sample in samples_list:
#             # we should have a list of T edge index tensors, T=num_timesteps
#             timesteps = len(sample.edge_indices)
#             num_timesteps = timesteps if timesteps > num_timesteps else num_timesteps
#         ## num_timesteps = len(samples_list[0].edge_index)
#
#         batched_graphs_t = []
#         for t in range(num_timesteps):
#             graphs_t = []
#             for i in range(num_sequences):
#                 try:
#                    graph_i = samples_list[i]
#                    # x = graph_i.x[t]
#                    # edge_index = graph_i.edge_index[t]
#                    # edge_attr = graph_i.edge_attr[t]
#                    # pos = graph_i.pos[t]
#                    # mask = graph_i.mask[t]
#                    # graph_it = Data(x=x,edge_index=edge_index,edge_attr=edge_attr,pos=pos,mask=mask)
#                    # graph_it = Data(**{k:v[t] for k,v in graph_i.to_dict().items() if type(v) in [torch.tensor, list]})
#                    graphs_t.append(graph_i.__get_item__(t))
#                 except IndexError:
#                    pass
#                    # TODO: do we need to do some sort of padding?
#                    # e.g.
#                    # g = Data(edge_index = torch.empty((2,0), dtype=torch.int64),
#                    #          x = torch.tensor(...),
#                    #          y = torch.tensor(...),
#                    #          mask = torch.tensor(...))
#                    # graphs_t.append(g)
#             batched_graphs_t.append(Batch.from_data_list(graphs_t))
#         return batched_graphs_t
#
#     def _get_loader(self, indices, **kwargs):
#         shuffle = kwargs.pop("shuffle", False)
#         batch_size = kwargs.pop("batch_size", 1)
#
#         dataset = self._get_dataset(**kwargs)
#         dataset = Subset(dataset, indices)
#
#         assert self.exp_seed is not None, 'DataLoader seed has not been specified! Is this a bug?'
#         kwargs['worker_init_fn'] = lambda worker_id: seed_worker(worker_id, self.exp_seed)
#
#         if shuffle is True:
#             sampler = RandomSampler(dataset)
#             dataloader = DataLoader(dataset, sampler=sampler,
#                                     num_workers=self.num_workers,
#                                     pin_memory=self.pin_memory,
#                                     batch_size=batch_size,
#                                     collate_fn=MultipleGraphSequenceDataProvider.collate_fn,
#                                     **kwargs)
#         else:
#             dataloader = DataLoader(dataset, shuffle=False,
#                                     num_workers=self.num_workers,
#                                     pin_memory=self.pin_memory,
#                                     batch_size=batch_size,
#                                     collate_fn=MultipleGraphSequenceDataProvider.collate_fn,
#                                     **kwargs)
#
#         return


import random

import numpy as np
import torch
from sklearn.model_selection import StratifiedKFold, StratifiedShuffleSplit, ShuffleSplit, KFold
from torch_geometric.utils import negative_sampling, to_undirected, to_dense_adj, add_self_loops

import pydgn.data.dataset
from pydgn.data.dataset import OGBGDatasetInterface
from pydgn.experiment.util import s2c


def to_lower_triangular(edge_index: torch.Tensor):
    r"""
    Transform Pytorch Geometric undirected edge index into its "lower triangular counterpart"
    """
    row, col = edge_index
    lower_tri_mask = row > col
    row, col = row[lower_tri_mask], col[lower_tri_mask]
    lower_tri_edge_index = torch.cat((row.unsqueeze(0), col.unsqueeze(0)), dim=0)
    return lower_tri_edge_index


class Fold:
    r"""
    Simple class that stores training, validation, and test indices.

    Args:
        train_idxs (Union[list, tuple]): training indices
        val_idxs (Union[list, tuple]): validation indices. Default is ``None``
        test_idxs (Union[list, tuple]): test indices. Default is ``None``
    """
    def __init__(self, train_idxs, val_idxs=None, test_idxs=None):
        self.train_idxs = train_idxs
        self.val_idxs = val_idxs
        self.test_idxs = test_idxs


class InnerFold(Fold):
    r"""
    Simple extension of the Fold class that returns a dictionary with training and validation indices (model selection).
    """
    def todict(self) -> dict:
        r"""
        Creates a dictionary with the training/validation indices.

        Returns:
            a dict with keys ``['train', 'val']`` associated with the respective indices
        """
        return {"train": self.train_idxs, "val": self.val_idxs}


class OuterFold(Fold):
    r"""
    Simple extension of the Fold class that returns a dictionary with training and test indices (risk assessment)
    """
    def todict(self) -> dict:
        r"""
        Creates a dictionary with the training/validation/test indices.

        Returns:
            a dict with keys ``['train', 'val', 'test']`` associated with the respective indices
        """
        return {"train": self.train_idxs, "val": self.val_idxs, "test": self.test_idxs}


class NoShuffleTrainTestSplit:
    r"""
    Class that implements a very simple training/test split. Can be used to further split training data into training
    and validation.

    Args:
        test_ratio: percentage of data to use for evaluation.
    """
    def __init__(self, test_ratio):
        self.test_ratio = test_ratio

    # Leave the arguments as they are. The parameter `y` is needed to implement the same interface as sklearn
    def split(self, idxs, y=None):
        n_samples = len(idxs)
        n_test = int(n_samples * self.test_ratio)
        n_train = n_samples - n_test
        train_idxs = np.arange(n_train)
        test_idxs = np.arange(n_train, n_train + n_test)
        return [(train_idxs, test_idxs)]


class Splitter:
    r"""
    Class that generates and stores the data splits at dataset creation time.

    Args:
        n_outer_folds (int): number of outer folds (risk assessment). 1 means hold-out, >1 means k-fold
        n_inner_folds (int): number of inner folds (model selection). 1 means hold-out, >1 means k-fold
        seed (int): random seed for reproducibility (on the same machine)
        stratify (bool): whether to apply stratification or not (should be true for classification tasks)
        shuffle (bool): whether to apply shuffle or not
        inner_val_ratio  (float): percentage of validation set for hold_out model selection. Default is ``0.1``
        outer_val_ratio  (float): percentage of validation set for hold_out model assessment (final training runs). Default is ``0.1``
        test_ratio  (float): percentage of test set for hold_out model assessment. Default is ``0.1``
    """
    def __init__(self,
                 n_outer_folds: int,
                 n_inner_folds: int,
                 seed: int,
                 stratify: bool,
                 shuffle: bool,
                 inner_val_ratio: float=0.1,
                 outer_val_ratio: float=0.1,
                 test_ratio: float=0.1):
        self.outer_folds = []
        self.inner_folds = []
        self.stratify = stratify
        self.shuffle = shuffle

        self.n_outer_folds = n_outer_folds
        self.n_inner_folds = n_inner_folds
        self.seed = seed

        self.inner_val_ratio = inner_val_ratio
        self.outer_val_ratio = outer_val_ratio
        self.test_ratio = test_ratio

    def get_graph_targets(self, dataset: pydgn.data.dataset.DatasetInterface) -> (bool, np.ndarray):
        r"""
        Reads the entire dataset and returns the targets.

        Args:
            dataset (:class:`~pydgn.data.dataset.DatasetInterface`): the dataset

        Returns:
            a tuple of two elements. The first element is a boolean, which is ``True`` if target values exist or an
            exception has not been thrown. The second value holds the actual targets or ``None``, depending on the
            first boolean value.
        """
        try:
            targets = np.array([d.y.item() for d in dataset])
            return True, targets
        except Exception:
            return False, None

    @classmethod
    def load(cls, path: str):
        r"""
        Loads the data splits from disk.

        Args:
            :param path: the path of the yaml file with the splits

        Returns:
            a :class:`~pydgn.data.splitter.Splitter` object
        """
        splits = torch.load(path)

        splitter_classname = splits.get("splitter_class", "Splitter")
        splitter_class = s2c(splitter_classname)

        splitter_args = splits.get("splitter_args")
        splitter = splitter_class(**splitter_args)

        assert splitter.n_outer_folds == len(splits["outer_folds"])
        assert splitter.n_inner_folds == len(splits["inner_folds"][0])

        for fold_data in splits["outer_folds"]:
            splitter.outer_folds.append(
                OuterFold(fold_data["train"], val_idxs=fold_data["val"], test_idxs=fold_data["test"]))

        for inner_split in splits["inner_folds"]:
            inner_split_data = []
            for fold_data in inner_split:
                inner_split_data.append(InnerFold(fold_data["train"], val_idxs=fold_data["val"]))
            splitter.inner_folds.append(inner_split_data)

        return splitter

    def _get_splitter(self, n_splits: int, stratified: bool, eval_ratio: float):
        r"""
        Instantiates the appropriate splitter to use depending on the situation.

        Args:
              n_splits (int): the number of different splits to create
              stratified (bool): whether or not to perform stratification. **Works with graph classification tasks only!**
              eval_ratio (float): the amount of evaluation (validation/test) data to use in case ``n_splits==1`` (i.e., hold-out data split)

        Returns:
            a :class:`~pydgn.data.splitter.Splitter` object
        """
        if n_splits == 1:
            if not self.shuffle:
                assert stratified == False, "Stratified not implemented when shuffle is False"
                splitter = NoShuffleTrainTestSplit(test_ratio=eval_ratio)
            else:
                if stratified:
                    splitter = StratifiedShuffleSplit(n_splits, test_size=eval_ratio, random_state=self.seed)
                else:
                    splitter = ShuffleSplit(n_splits, test_size=eval_ratio, random_state=self.seed)
        elif n_splits > 1:
            if stratified:
                splitter = StratifiedKFold(n_splits, shuffle=self.shuffle, random_state=self.seed)
            else:
                splitter = KFold(n_splits, shuffle=self.shuffle, random_state=self.seed)
        else:
            raise ValueError(f"'n_splits' must be >=1, got {n_splits}")

        return splitter

    def split(self, dataset: pydgn.data.dataset.DatasetInterface, targets: np.ndarray=None):
        r"""
        Computes the splits and stores them in the list fields ``self.outer_folds`` and ``self.inner_folds``.

        Args:
            dataset (:class:`~pydgn.data.dataset.DatasetInterface`): the Dataset object
            targets (np.ndarray]): targets used for stratification. Default is ``None``
        """
        idxs = range(len(dataset))

        stratified = self.stratify
        outer_idxs = np.array(idxs)

        outer_splitter = self._get_splitter(
            n_splits=self.n_outer_folds,
            stratified=stratified,
            eval_ratio=self.test_ratio)  # This is the true test (outer test)

        for train_idxs, test_idxs in outer_splitter.split(outer_idxs, y=targets):

            assert set(train_idxs) == set(outer_idxs[train_idxs])
            assert set(test_idxs) == set(outer_idxs[test_idxs])

            inner_fold_splits = []
            inner_idxs = outer_idxs[train_idxs]  # equals train_idxs because outer_idxs was ordered
            inner_targets = targets[train_idxs] if targets is not None else None

            inner_splitter = self._get_splitter(
                n_splits=self.n_inner_folds,
                stratified=stratified,
                eval_ratio=self.inner_val_ratio)  # The inner "test" is, instead, the validation set

            for inner_train_idxs, inner_val_idxs in inner_splitter.split(inner_idxs, y=inner_targets):
                inner_fold = InnerFold(train_idxs=inner_idxs[inner_train_idxs].tolist(),
                                       val_idxs=inner_idxs[inner_val_idxs].tolist())
                inner_fold_splits.append(inner_fold)

                # False if empty
                assert not bool(set(inner_train_idxs) & set(inner_val_idxs) & set(test_idxs))
                assert not bool(set(inner_idxs[inner_train_idxs]) & set(inner_idxs[inner_val_idxs]) & set(test_idxs))

            self.inner_folds.append(inner_fold_splits)

            # Obtain outer val from outer train in an holdout fashion
            outer_val_splitter = self._get_splitter(n_splits=1, stratified=stratified,
                                                    eval_ratio=self.outer_val_ratio)
            outer_train_idxs, outer_val_idxs = list(outer_val_splitter.split(inner_idxs, y=inner_targets))[0]

            # False if empty
            assert not bool(set(outer_train_idxs) & set(outer_val_idxs) & set(test_idxs))
            assert not bool(set(outer_train_idxs) & set(outer_val_idxs) & set(test_idxs))
            assert not bool(set(inner_idxs[outer_train_idxs]) & set(inner_idxs[outer_val_idxs]) & set(test_idxs))

            np.random.shuffle(outer_train_idxs)
            np.random.shuffle(outer_val_idxs)
            np.random.shuffle(test_idxs)
            outer_fold = OuterFold(train_idxs=inner_idxs[outer_train_idxs].tolist(),
                                   val_idxs=inner_idxs[outer_val_idxs].tolist(),
                                   test_idxs=outer_idxs[test_idxs].tolist())
            self.outer_folds.append(outer_fold)

    def _splitter_args(self) -> dict:
        r"""
        Returns a dict with all the splitter's arguments for subsequent re-loading at experiment time.

        Returns:
            a dict containing all splitter's arguments.
        """
        return {
            "n_outer_folds": self.n_outer_folds,
            "n_inner_folds": self.n_inner_folds,
            "seed": self.seed,
            "stratify": self.stratify,
            "shuffle": self.shuffle,
            "inner_val_ratio": self.inner_val_ratio,
            "outer_val_ratio": self.outer_val_ratio,
            "test_ratio": self.test_ratio,
        }

    def save(self, path: str):
        r"""
        Saves the split as a dictionary into a ``torch`` file. The arguments of the dictionary are
        * seed (int)
        * splitter_class (str)
        * splitter_args (dict)
        * outer_folds (list of dicts)
        * inner_folds (list of lists of dicts)

        Args:
            path (str): filepath where to save the object
        """
        print("Saving splits on disk...")

        # save split class name
        module = self.__class__.__module__
        splitter_class = self.__class__.__qualname__
        if module is not None and module != "__builtin__":
            splitter_class = module + "." + splitter_class

        savedict = {"seed": self.seed,
                    "splitter_class": splitter_class,
                    "splitter_args": self._splitter_args(),
                    "outer_folds": [o.todict() for o in self.outer_folds],
                    "inner_folds": []}

        for inner_split in self.inner_folds:  # len(self.inner_folds) == # of **outer** folds
            savedict["inner_folds"].append([i.todict() for i in inner_split])
        torch.save(savedict, path)
        print("Done.")


class TemporalSplitter(Splitter):
    r"""
    Reads the entire dataset and returns the targets. In this case, each sample in the dataset represents
    a temporal graph, so we get the classification value at the last time step. Use this method to stratify a dataset
    of multiple temporal graphs.

    Args:
        dataset (:class:`~pydgn.data.dataset.DatasetInterface`): the dataset

    Returns:
        a tuple of two elements. The first element is a boolean, which is ``True`` if target values exist or an
        exception has not been thrown. The second value holds the actual targets or ``None``, depending on the
        first boolean value.
    """
    def get_graph_targets(self, dataset):
        try:
            targets = np.array([d.targets[-1].item() for d in dataset])
            return True, targets
        except Exception:
            return False, None


class OGBGSplitter(Splitter):

    def split(self, dataset: OGBGDatasetInterface, targets=None):
        assert self.n_outer_folds == 1 and self.n_inner_folds == 1, "OGBGSplitter assumes you want to use the same splits as in the original dataset!"
        original_splits = dataset.get_idx_split()

        outer_train_indices = original_splits['train'].numpy().tolist()
        outer_valid_indices = original_splits['valid'].numpy().tolist()
        outer_test_indices = original_splits['test'].numpy().tolist()

        np.random.shuffle(outer_train_indices)
        np.random.shuffle(outer_valid_indices)
        np.random.shuffle(outer_test_indices)

        inner_fold_splits = []
        inner_fold = InnerFold(train_idxs=outer_train_indices, val_idxs=outer_valid_indices)
        inner_fold_splits.append(inner_fold)
        self.inner_folds.append(inner_fold_splits)

        outer_fold = OuterFold(train_idxs=outer_train_indices, val_idxs=outer_valid_indices,
                               test_idxs=outer_test_indices)
        self.outer_folds.append(outer_fold)


class SingleGraphSequenceSplitter(TemporalSplitter):
    """
    Class for dynamic graphs that generates the splits at dataset creation time. It assumes that there is a single graph
    sequence, so the split happens on time steps. What is more, `n_inner_folds` here will create an inner K-fold CV
    split for model selection where, however, the training/validation split will not change (because there is no way
    to split time steps in a different way). This allows for different initializations of the same model, evaluating
    the avg performance on the VL set.
    """
    def __init__(self,
                 n_outer_folds: int,
                 n_inner_folds: int,
                 seed: int,
                 stratify: bool,
                 shuffle: bool,
                 inner_val_ratio: float=0.1,
                 outer_val_ratio: float=0.1,
                 test_ratio: float=0.1):
        super().__init__(n_outer_folds, n_inner_folds, seed, stratify, shuffle,
                         inner_val_ratio, outer_val_ratio, test_ratio)
        assert n_outer_folds == 1, "SingleGraphSequenceSplitter does not allow more than one outer fold, as the graph sequence is unique."
        assert not shuffle, "SingleGraphSequenceSplitter does not allow to shuffle the time steps of the unique graph sequence."
        assert not stratify, "You cannot stratify a single graph sequence."

    def split(self, dataset: pydgn.data.dataset.DatasetInterface, targets: np.ndarray=None):
        r"""
        Computes the splits and stores them in the list fields ``self.outer_folds`` and ``self.inner_folds``.

        Args:
            dataset (:class:`~pydgn.data.dataset.DatasetInterface`): the Dataset object
            targets (np.ndarray]): targets used for stratification. Default is ``None``
        """
        idxs = range(len(dataset))
        print(f'Number of snapshots: {len(dataset)}')

        outer_splitter = self._get_splitter(
            n_splits=self.n_outer_folds,
            stratified=self.stratify,
            eval_ratio=self.test_ratio)  # This is the true test (outer test)

        outer_idxs = np.array(idxs)
        train_idxs, test_idxs = outer_splitter.split(outer_idxs, y=targets)[0]

        assert set(train_idxs) == set(outer_idxs[train_idxs])
        assert set(test_idxs) == set(outer_idxs[test_idxs])

        inner_fold_splits = []
        inner_idxs = outer_idxs[train_idxs]  # equals train_idxs because outer_idxs was ordered

        inner_splitter = NoShuffleTrainTestSplit(test_ratio=self.inner_val_ratio)

        # Repeat the very same inner split for n_inner_folds, as there is no way to split differently.
        # This allows for different initializations of the same model, evaluating the avg performance on the VL set.
        inner_train_idxs, inner_val_idxs =  list(inner_splitter.split(inner_idxs, y=None))[0]
        for _ in range(self.n_inner_folds):
            inner_fold = InnerFold(train_idxs=inner_idxs[inner_train_idxs].tolist(),
                                   val_idxs=inner_idxs[inner_val_idxs].tolist())
            inner_fold_splits.append(inner_fold)

        self.inner_folds.append(inner_fold_splits)

        # Obtain outer val from outer train in an holdout fashion
        outer_val_splitter = NoShuffleTrainTestSplit(test_ratio=self.outer_val_ratio)
        outer_train_idxs, outer_val_idxs = list(outer_val_splitter.split(inner_idxs, y=None))[0]

        # False if empty
        assert not bool(set(inner_train_idxs) & set(inner_val_idxs) & set(test_idxs))
        assert not bool(set(inner_idxs[inner_train_idxs]) & set(inner_idxs[inner_val_idxs]) & set(test_idxs))
        assert not bool(set(outer_train_idxs) & set(outer_val_idxs) & set(test_idxs))
        assert not bool(set(outer_train_idxs) & set(outer_val_idxs) & set(test_idxs))
        assert not bool(set(inner_idxs[outer_train_idxs]) & set(inner_idxs[outer_val_idxs]) & set(test_idxs))

        outer_fold = OuterFold(train_idxs=inner_idxs[outer_train_idxs].tolist(),
                               val_idxs=inner_idxs[outer_val_idxs].tolist(),
                               test_idxs=outer_idxs[test_idxs].tolist())
        self.outer_folds.append(outer_fold)


class LinkPredictionSingleGraphSplitter(Splitter):
    r"""
    Class that inherits from :class:`~pydgn.data.splitter.Splitter` and computes link splits for link classification tasks.
    **IMPORTANT:** This class implements **bootstrapping** rather than k-fold cross-validation, so different
    outer test sets may have overlapping indices.

    **Does not support edge attributes at the moment.**

    Args:
        n_outer_folds (int): number of outer folds (risk assessment). 1 means hold-out, >1 means k-fold
        n_inner_folds (int): number of inner folds (model selection). 1 means hold-out, >1 means k-fold
        seed (int): random seed for reproducibility (on the same machine)
        stratify (bool): whether to apply stratification or not (should be true for classification tasks)
        shuffle (bool): whether to apply shuffle or not
        inner_val_ratio  (float): percentage of validation set for hold_out model selection. Default is ``0.1``
        outer_val_ratio  (float): percentage of validation set for hold_out model assessment (final training runs). Default is ``0.1``
        test_ratio  (float): percentage of test set for hold_out model assessment. Default is ``0.1``
        undirected (bool): whether or not the graph is undirected. Default is ``False``
        avoid_opposite_negative_edges (bool): whether or not to **avoid** creating negative edges that are opposite to existing edges Default is ``True``
        run_checks (bool): whether or not to run correctness checks. Creates a full adjacency matrix, can be memory intensive.
    """
    def __init__(self,
                 n_outer_folds: int,
                 n_inner_folds: int,
                 seed: int,
                 stratify: bool,
                 shuffle: bool,
                 inner_val_ratio: float=0.1,
                 outer_val_ratio: float=0.1,
                 test_ratio: float=0.1,
                 undirected: bool=False,
                 avoid_opposite_negative_edges: bool=True,
                 run_checks=False):
        super().__init__(n_outer_folds,
                         n_inner_folds,
                         seed,
                         stratify,
                         shuffle,
                         inner_val_ratio,
                         outer_val_ratio,
                         test_ratio)
        self.run_checks = run_checks
        self.undirected = undirected
        self.avoid_opposite_negative_edges = avoid_opposite_negative_edges

    @staticmethod
    def _run_checks(A, pos_train_edges, neg_train_edges, pos_val_edges, neg_val_edges, pos_test_edges,
                    neg_test_edges):
        print(
            f'Inner splits: {pos_train_edges.shape}, {neg_train_edges.shape}, {pos_val_edges.shape}, {neg_val_edges.shape}')

        # First: check positive edges perfectly reconstruct adjacency matrix
        N = A.shape[0]
        # TEST: restore A with partitioned positive links
        A_pos = to_dense_adj(pos_train_edges, max_num_nodes=N)[0, :, :]
        A_pos += to_dense_adj(pos_val_edges, max_num_nodes=N)[0, :, :]
        A_pos += to_dense_adj(pos_test_edges, max_num_nodes=N)[0, :, :]

        # It is also a way to check that positive edges do not overlap
        # If I do not call to_undirected to val and test edges when undirected==True, this won't hold
        # assert torch.all(A == A_pos)

        # Second: check edges do not overlap
        # Can be done by checking the sum of positive and negative adj matrices holds values smaller or equal than 1.

        # TEST: check negative links are separate from positive links
        A_neg = to_dense_adj(neg_train_edges, max_num_nodes=N)[0, :, :]
        A_neg += to_dense_adj(neg_val_edges, max_num_nodes=N)[0, :, :]
        A_neg += to_dense_adj(neg_test_edges, max_num_nodes=N)[0, :, :]

        assert torch.all((A_neg + A_pos) <= 1.)

    def train_val_test_edge_split(self, edge_index, edge_attr, val_ratio, test_ratio, num_nodes):
        r"""
        Sample training/validation/test edges at random.
        """
        if self.undirected:
            edge_index = to_lower_triangular(edge_index)

        no_edges = edge_index.shape[1]
        permutation = list(range(no_edges))
        random.shuffle(permutation)

        no_val_elements, no_test_elements = int(no_edges * val_ratio), int(no_edges * test_ratio)

        train_elements = permutation[no_val_elements + no_test_elements:]
        val_elements = permutation[:no_val_elements]
        test_elements = permutation[no_val_elements:no_val_elements + no_test_elements]

        assert set(train_elements).intersection(set(test_elements)) == set()
        assert set(train_elements).intersection(set(val_elements)) == set()
        assert set(val_elements).intersection(set(test_elements)) == set()

        train_edges = edge_index[:, train_elements]
        val_edges = edge_index[:, val_elements]
        test_edges = edge_index[:, test_elements]

        # Restore training edges as undirected for training
        # and double validation and test edges to maintain proportions
        if self.undirected:
            train_edges = to_undirected(train_edges, num_nodes=num_nodes) if train_edges.shape[1] > 0 else train_edges
            val_edges = to_undirected(val_edges, num_nodes=num_nodes) if val_edges.shape[1] > 0 else val_edges
            test_edges = to_undirected(test_edges, num_nodes=num_nodes) if test_edges.shape[1] > 0 else test_edges

        # Use the permuted elements to recover the associated train/val/test edge attributes
        # TODO: If undirected, train_attr should be doubled according to train edges and the logic of to_undirected!
        train_attr = None  # edge_attr[train_elements].tolist() if edge_attr is not None else None
        val_attr = None  # edge_attr[val_elements].tolist() if edge_attr is not None else None
        test_attr = None  # edge_attr[test_elements].tolist() if edge_attr is not None else None

        return train_edges, train_attr, val_edges, val_attr, test_edges, test_attr

    def split(self, dataset: pydgn.data.dataset.DatasetInterface, targets: np.ndarray=None):
        r"""
        Computes the splits and stores them in the list fields ``self.outer_folds`` and ``self.inner_folds``.
        Links are selected at random: this means outer test folds will overlap almost surely with if test_ratio is 10% of the total samples.
        The recommended procedure here is to use the outer folds to do bootstrapping rather than k-fold cross-validation.
        Idea taken from: https://arxiv.org/pdf/1811.05868.pdf

        Args:
            dataset (:class:`~pydgn.data.dataset.DatasetInterface`): the Dataset object
            targets (np.ndarray]): targets used for stratification. Default is ``None``
        """
        random.seed(self.seed)
        assert len(dataset) == 1, f"LinkPredictionSingleGraphSplitter works on single graph dataset only! Here we have {len(dataset)}"
        edge_index = dataset.data.edge_index
        edge_attr = dataset.data.edge_attr
        num_nodes = dataset.data.x.shape[0]

        if self.undirected and self.run_checks:
            A = to_dense_adj(edge_index)[0]
            assert torch.all(A == A.transpose(1, 0)), "Passed undirected == True with a non-symmetric adj matrix"
        elif not self.undirected and self.run_checks:
            A = to_dense_adj(edge_index)[0]
            assert torch.any(A != A.transpose(1, 0)), "Passed undirected == False with a symmetric adj matrix"


        inner_val_ratio = self.inner_val_ratio if self.n_inner_folds == 1 else float(1 / self.n_inner_folds)
        outer_val_ratio = self.outer_val_ratio if self.n_inner_folds == 1 else float(1 / self.n_inner_folds)
        test_ratio = self.test_ratio if self.n_outer_folds == 1 else float(1 / self.n_outer_folds)

        # Used by negative sampling later
        original_edge_index = edge_index

        # Add self loops to prevent negative edges from being self loops
        original_edge_index_plus_selfloops, _ = add_self_loops(original_edge_index)

        for outer_k in range(self.n_outer_folds):
            print(f'Processing splits for outer fold n {outer_k + 1}')

            # Create positive train and test edges for outer fold
            print(f'Split positive edges from edge index of shape {original_edge_index.shape[1]}')
            pos_train_edges, train_attr, outer_pos_val_edges, outer_val_attr, pos_test_edges, test_attr = self.train_val_test_edge_split(
                original_edge_index, edge_attr, outer_val_ratio, test_ratio, num_nodes)

            ###
            # Generate NEGATIVE training/validation/test edges for each fold using the original edge_index
            ###

            # We generate directed negative edges of the same number as true edges
            num_neg_samples = original_edge_index.shape[1]

            # Note: negative edges are directed, but can be considered as undirected if the loss/score is symmetric
            if self.avoid_opposite_negative_edges:
                # We generate symmetric edges to avoid sampling an opposite edge as negative
                # Remember that negative sampling without additional args samples directed edges, so we are good
                negative_edge_index = negative_sampling(to_undirected(original_edge_index_plus_selfloops),
                                                        num_nodes=num_nodes, num_neg_samples=num_neg_samples)
            else:
                negative_edge_index = negative_sampling(original_edge_index_plus_selfloops, num_nodes=num_nodes,
                                                        num_neg_samples=num_neg_samples)

            # Negative edges are directed, and it does not make sense to double them.
            tmp_undirected = self.undirected
            self.undirected = False
            neg_train_edges, _, outer_neg_val_edges, _, neg_test_edges, _ = self.train_val_test_edge_split(
                negative_edge_index, None, outer_val_ratio, test_ratio, num_nodes=num_nodes)
            self.undirected = tmp_undirected

            if self.run_checks:
                print(
                    f'Outer splits: {pos_train_edges.shape}, {neg_train_edges.shape}, {outer_pos_val_edges.shape}, {outer_neg_val_edges.shape}, {pos_test_edges.shape}, {neg_test_edges.shape}')

            inner_fold_splits = []
            for inner_k in range(self.n_inner_folds):

                # Do not need to create additional negative edges here! Just split what we already have

                # Create positive train and validation edges for outer fold
                inner_pos_train_edges, inner_train_attr, inner_pos_val_edges, inner_val_attr, _, _ = self.train_val_test_edge_split(
                    pos_train_edges, train_attr, inner_val_ratio, 0., num_nodes=num_nodes)

                # Negative edges are directed, and it does not make sense to double them.
                tmp_undirected = self.undirected
                self.undirected = False
                # Create negative train and validation edges for outer fold
                inner_neg_train_edges, _, inner_neg_val_edges, _, _, _ = self.train_val_test_edge_split(
                    neg_train_edges, None, inner_val_ratio, 0., num_nodes=num_nodes)
                self.undirected = tmp_undirected

                inner_fold = InnerFold(train_idxs=(inner_pos_train_edges.tolist(),
                                                   inner_train_attr,
                                                   inner_neg_train_edges.tolist()),
                                       val_idxs=(inner_pos_val_edges.tolist(),
                                                 inner_val_attr,
                                                 inner_neg_val_edges.tolist()))
                inner_fold_splits.append(inner_fold)

                if self.run_checks:
                    print("Running additional checks: disable this option after testing the function")
                    self._run_checks(A, inner_pos_train_edges, inner_neg_train_edges, inner_pos_val_edges,
                                     inner_neg_val_edges, pos_test_edges, neg_test_edges)

            # Finally, check we are not creating positive links for some mistake
            if self.run_checks:
                A = to_dense_adj(edge_index)[0]
                neg_A = to_dense_adj(negative_edge_index, max_num_nodes=A.shape[0])[0, :, :]
                assert torch.all(A + neg_A <= 1.)

            self.inner_folds.append(inner_fold_splits)

            outer_fold = OuterFold(train_idxs=(pos_train_edges.tolist(),
                                               train_attr,
                                               neg_train_edges.tolist()),
                                   val_idxs=(outer_pos_val_edges.tolist(),
                                             outer_val_attr,
                                             outer_neg_val_edges.tolist()),
                                   test_idxs=(pos_test_edges.tolist(),
                                              test_attr,
                                              neg_test_edges.tolist()))
            self.outer_folds.append(outer_fold)

    def _splitter_args(self):
        splitter_args = super()._splitter_args()
        splitter_args.update({"undirected": self.undirected,
                              "avoid_opposite_negative_edges": self.avoid_opposite_negative_edges})
        return splitter_args



import random
from typing import Callable, Tuple, List

import numpy as np
import torch

from pydgn.data.provider import DataProvider
from pydgn.evaluation.config import Config
from pydgn.evaluation.util import return_class_and_args
from pydgn.experiment.util import s2c
from pydgn.log.logger import Logger
from pydgn.model.interface import ModelInterface, ReadoutInterface
from pydgn.static import DEFAULT_ENGINE_CALLBACK
from pydgn.training.engine import TrainingEngine


class Experiment:
    r"""
    Class that handles a single experiment.

    Args:
        model_configuration (dict): the dictionary holding the experiment-specific configuration
        exp_path (str): path to the experiment folder
        exp_seed (int): the experiment's seed to use
    """
    def __init__(self, model_configuration: dict, exp_path: str, exp_seed: int):
        self.model_config = Config(model_configuration)
        self.exp_path = exp_path
        self.exp_seed = exp_seed
        # Set seed here to aid reproducibility
        np.random.seed(self.exp_seed)
        torch.manual_seed(self.exp_seed)
        torch.cuda.manual_seed(self.exp_seed)
        random.seed(self.exp_seed)

        # torch.use_deterministic_algorithms(True) for future versions of Pytorch
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

    def _return_class_and_args(config: Config, key: str) -> Tuple[Callable[...,object],
                                                                        dict]:
        r"""
        Returns the class and arguments associated to a specific key in the configuration file.

        Args:
            config: the configuration dictionary
            key: a string representing a particular class in the configuration dictionary

        Returns:
            a tuple (class, dict of arguments), or (None, None) if the key is not present in the config dictionary
        """
        if key not in config or config[key] is None:
            return None, None
        elif isinstance(config[key], str):
            return s2c(config[key]), {}
        elif isinstance(config[key], dict):
            return s2c(config[key]['class_name']), config[key]['args'] if 'args' in config[key] else {}
        else:
            raise NotImplementedError('Parameter has not been formatted properly')

    def _create_model(self,
                      dim_node_features: int,
                      dim_edge_features: int,
                      dim_target: int,
                      readout_classname: str,
                      config: Config) -> ModelInterface:
        r"""
        Instantiates a model that implements the :class:`~pydgn.model.model.ModelInterface` interface

        Args:
            dim_node_features (int): number of node features
            dim_edge_features (int): number of edge features
            dim_target (int): target dimension
            readout_classname (str): string containing the model's class
            config (:class:`~pydgn.evaluation.config.Config`): the configuration dictionary

        Returns:
            a model that implements the :class:`~pydgn.model.model.ModelInterface` interface
        """
        model = s2c(config['model'])(dim_node_features=dim_node_features,
                                             dim_edge_features=dim_edge_features,
                                             dim_target=dim_target,
                                             readout_class=s2c(readout_classname)
                                             if readout_classname is not None else None,
                                             config=config)

        # move to device
        model.to(self.model_config.device)  # model .to() may not return anything
        return model

    def create_supervised_model(self,
                                dim_node_features: int,
                                dim_edge_features: int,
                                dim_target: int) -> ModelInterface:
        r"""
        Instantiates a **supervised** model that implements the :class:`~pydgn.model.model.ModelInterface` interface,
        using the ``supervised_config`` field in the configuration file.

        Args:
            dim_node_features (int): number of node features
            dim_edge_features (int): number of edge features
            dim_target (int): target dimension

        Returns:
            a model that implements the :class:`~pydgn.model.model.ModelInterface` interface
        """
        readout_classname = self.model_config.supervised_config['readout'] \
            if 'readout' in self.model_config.supervised_config else None
        return self._create_model(dim_node_features, dim_edge_features, dim_target, readout_classname,
                                  self.model_config.supervised_config)

    def create_supervised_readout(self,
                                dim_node_features: int,
                                dim_edge_features: int,
                                dim_target: int) -> ReadoutInterface:
        r"""
        Instantiates an **supervised** readout that implements the :class:`~pydgn.model.model.ReadoutInterface` interface,
        using the ``supervised_config`` field in the configuration file.

        Args:
            dim_node_features (int): number of node features
            dim_edge_features (int): number of edge features
            dim_target (int): target dimension

        Returns:
            a model that implements the :class:`~pydgn.model.model.ReadoutInterface` interface
        """
        return s2c(self.model_config.supervised_config['readout'])(dim_node_features=dim_node_features,
                                                                     dim_edge_features=dim_edge_features,
                                                                     dim_target=dim_target,
                                                                     config=self.model_config.supervised_config)

    def create_unsupervised_model(self,
                                  dim_node_features: int,
                                  dim_edge_features: int,
                                  dim_target: int) -> ModelInterface:
        r"""
        Instantiates an **unsupervised** model that implements the :class:`~pydgn.model.model.ModelInterface` interface,
        using the ``unsupervised_config`` field in the configuration file.

        Args:
            dim_node_features (int): number of node features
            dim_edge_features (int): number of edge features
            dim_target (int): target dimension

        Returns:
            a model that implements the :class:`~pydgn.model.model.ModelInterface` interface
        """
        readout_classname = self.model_config.unsupervised_config['readout'] \
            if 'readout' in self.model_config.unsupervised_config else None
        return self._create_model(dim_node_features, dim_edge_features, dim_target, readout_classname,
                                  self.model_config.unsupervised_config)

    def create_incremental_model(self,
                                 dim_node_features: (int),
                                 dim_edge_features: (int),
                                 dim_target: (int),
                                 depth: (int),
                                 prev_outputs_to_consider: List[int]) -> ModelInterface:
        r"""
        Instantiates a layer of an incremental architecture. It assumes the config file has a field ``layer_config``
        and another ``layer_config.arbitrary_function_config`` that holds any kind of information for the arbitrary
        function of an incremental architecture

        Args:
            dim_node_features: input node features
            dim_edge_features: input edge features
            dim_target: target size
            depth: current depth of the architecture
            prev_outputs_to_consider: A list of previous layers to consider, e.g. [1,2] means the last two previous layers.

        Returns:
            a layer of a model that implements the :class:`~pydgn.model.model.ModelInterface` interface
        """
        readout_classname = self.model_config.layer_config.get('readout', None)
        self.model_config.layer_config['depth'] = depth
        self.model_config.layer_config['prev_outputs_to_consider'] = prev_outputs_to_consider
        return self._create_model(dim_node_features, dim_edge_features, dim_target,
                                  readout_classname, self.model_config.layer_config)

    def _create_engine(self,
                        config: Config,
                        model: ModelInterface,
                        device: str,
                        evaluate_every: int,
                        reset_eval_model_hidden_state: bool) -> TrainingEngine:
        r"""
        Utility that instantiates the training engine. It looks for pre-defined fields in the configuration file,
        i.e. ``loss``, ``scorer``, ``optimizer``, ``scheduler``, ``gradient_clipper``, ``early_stopper`` and
        ``plotter``, all of which should be classes implementing the :class:`~pydgn.training.event.handler.EventHandler` interface

        Args:
            config (:class:`~pydgn.evaluation.config.Config`): the configuration dictionary
            model: the  model that needs be trained
            device (str): the string with the CUDA device to be used, or ``cpu``
            evaluate_every (int): number of epochs after which to log information
            reset_eval_model_hidden_state (bool): [temporal graph learning] Used when we want to reset the state after performing previous inference. It should be ``False`` when we are dealing with a single temporal graph sequence, because we don't want to reset the hidden state after processing the previous [training/validation] time steps.

        Returns:
            a :class:`~pydgn.training.engine.TrainingEngine` object
        """

        loss_class, loss_args = return_class_and_args(config, 'loss')
        loss = loss_class(use_as_loss=True, **loss_args) if loss_class is not None else None

        scorer_class, scorer_args = return_class_and_args(config, 'scorer')
        scorer = scorer_class(use_as_loss=False, **scorer_args) if scorer_class is not None else None

        optim_class, optim_args = return_class_and_args(config, 'optimizer')
        optimizer = optim_class(model=model, **optim_args) if optim_class is not None else None

        sched_class, sched_args = return_class_and_args(config, 'scheduler')
        if sched_args is not None:
            sched_args['optimizer'] = optimizer.optimizer
        scheduler = sched_class(**sched_args) if sched_class is not None else None
        # Remove the optimizer obj ow troubles when dumping the config file
        if sched_args is not None:
            sched_args.pop('optimizer', None)

        grad_clip_class, grad_clip_args = return_class_and_args(config, 'gradient_clipper')
        grad_clipper = grad_clip_class(**grad_clip_args) if grad_clip_class is not None else None

        early_stop_class, early_stop_args = return_class_and_args(config, 'early_stopper')
        early_stopper = early_stop_class(**early_stop_args) if early_stop_class is not None else None

        plot_class, plot_args = return_class_and_args(config, 'plotter')
        plotter = plot_class(exp_path=self.exp_path, **plot_args) if plot_class is not None else None

        store_last_checkpoint = config.get('checkpoint', False)
        engine_class, engine_args = return_class_and_args(config, 'engine')
        engine_callback = s2c(engine_args.get('engine_callback', DEFAULT_ENGINE_CALLBACK))

        engine = engine_class(engine_callback=engine_callback, model=model, loss=loss,
                                optimizer=optimizer, scorer=scorer, scheduler=scheduler,
                                early_stopper=early_stopper, gradient_clipper=grad_clipper,
                                device=device, plotter=plotter, exp_path=self.exp_path, evaluate_every=evaluate_every,
                                store_last_checkpoint=store_last_checkpoint,
                                reset_eval_model_hidden_state=reset_eval_model_hidden_state)
        return engine

    def create_supervised_engine(self, model: ModelInterface) -> TrainingEngine:
        r"""
        Instantiates the training engine by using the ``supervised_config`` key in the config file

        Args:
            model: the  model that needs be trained

        Returns:
            a :class:`~pydgn.training.engine.TrainingEngine` object
        """
        device = self.model_config.device
        evaluate_every = self.model_config.evaluate_every
        reset_eval_model_hidden_state = self.model_config.get('reset_eval_model_hidden_state', True)

        return self._create_engine(self.model_config.supervised_config, model, device,
                                   evaluate_every, reset_eval_model_hidden_state)

    def create_unsupervised_engine(self, model: ModelInterface) -> TrainingEngine:
        r"""
        Instantiates the training engine by using the ``unsupervised_config`` key in the config file

        Args:
            model: the  model that needs be trained

        Returns:
            a :class:`~pydgn.training.engine.TrainingEngine` object
        """
        device = self.model_config.device
        evaluate_every = self.model_config.evaluate_every
        reset_eval_model_hidden_state = self.model_config.get('reset_eval_model_hidden_state', True)
        return self._create_engine(self.model_config.unsupervised_config, model, device,
                                   evaluate_every, reset_eval_model_hidden_state)

    def create_incremental_engine(self, model: ModelInterface) -> TrainingEngine:
        r"""
        Instantiates the training engine by using the ``layer_config`` key in the config file

        Args:
            model: the  model that needs be trained

        Returns:
            a :class:`~pydgn.training.engine.TrainingEngine` object
        """
        device = self.model_config.device
        evaluate_every = self.model_config.evaluate_every
        reset_eval_model_hidden_state = self.model_config.get('reset_eval_model_hidden_state', True)
        return self._create_engine(self.model_config.layer_config, model, device,
                                   evaluate_every, reset_eval_model_hidden_state)

    def run_valid(self, dataset_getter, logger) -> Tuple[dict, dict]:
        r"""
        This function returns the training and validation results for a `model selection run`. **Do not attempt to load the set inside this method!**
        **If possible, rely on already available subclasses of this class**.

        Args:
            dataset_getter (:class:`~pydgn.data.provider.DataProvider`): a data provider
            logger (:class:`~pydgn.log.logger.Logger`): the logger

        Returns:
            a tuple of training and test dictionaries. Each dictionary has two keys:

            * ``LOSS`` (as defined in ``pydgn.static``)
            * ``SCORE`` (as defined in ``pydgn.static``)

            For instance, training_results[SCORE] is a dictionary itself with other fields to be used by the evaluator.
        """
        raise NotImplementedError('You must implement this function!')

    def run_test(self, dataset_getter: DataProvider, logger: Logger) -> Tuple[dict, dict, dict]:
        """
        This function returns the training, validation and test results for a `final run`. **Do not use the test to train the model nor for early stopping reasons!**
        **If possible, rely on already available subclasses of this class**.

        Args:
            dataset_getter (:class:`~pydgn.data.provider.DataProvider`): a data provider
            logger (:class:`~pydgn.log.logger.Logger`): the logger

        Returns:
            a tuple of training,validation,test dictionaries. Each dictionary has two keys:

            * ``LOSS`` (as defined in ``pydgn.static``)
            * ``SCORE`` (as defined in ``pydgn.static``)

            For instance, training_results[SCORE] is a dictionary itself with other fields to be used by the evaluator.
        """
        raise NotImplementedError('You must implement this function!')



from typing import Callable, Tuple, Optional, List

import torch
from torch_geometric.data import Batch


class ModelInterface(torch.nn.Module):
    r"""
    Provides the signature for any main model to be trained under PyDGN

    Args:
        dim_node_features (int): dimension of node features (according to the :class:`~pydgn.data.dataset.DatasetInterface` property)
        dim_edge_features (int): dimension of edge features (according to the :class:`~pydgn.data.dataset.DatasetInterface` property)
        dim_target (int): dimension of the target (according to the :class:`~pydgn.data.dataset.DatasetInterface` property)
        readout_class (Callable[...,:class:`torch.nn.Module`]): class of the module implementing the readout. This is optional, but useful to put different readouts to try in the config file
        config (dict): config dictionary containing all the necessary hyper-parameters plus additional information (if needed)
    """
    def __init__(self,
                 dim_node_features: int,
                 dim_edge_features: int,
                 dim_target: int,
                 readout_class: Callable[...,torch.nn.Module],
                 config: dict):
        super().__init__()
        self.dim_node_features = dim_node_features
        self.dim_edge_features = dim_edge_features
        self.dim_target = dim_target
        self.readout_class = readout_class
        self.config = config

    def forward(self, data: Batch) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[List[object]]]:
        r"""
        Performs a forward pass over a batch of graphs

        Args:
            data (:class:`torch_geometric.data.Batch`): a batch of graphs

        Returns:
            a tuple (model's output, [optional] node embeddings, [optional] additional outputs
        """
        raise NotImplementedError('You need to implement this method!')


class ReadoutInterface(torch.nn.Module):
    r"""
    Provides the signature for any readout to be trained under PyDGN

    Args:
        dim_node_features (int): dimension of node features (according to the :class:`~pydgn.data.dataset.DatasetInterface` property)
        dim_edge_features (int): dimension of edge features (according to the :class:`~pydgn.data.dataset.DatasetInterface` property)
        dim_target (int): dimension of the target (according to the :class:`~pydgn.data.dataset.DatasetInterface` property)
        config (dict): config dictionary containing all the necessary hyper-parameters plus additional information (if needed)
    """
    def __init__(self,
                 dim_node_features: int,
                 dim_edge_features: int,
                 dim_target: int,
                 config: dict):
        super().__init__()
        self.dim_node_features = dim_node_features
        self.dim_edge_features = dim_edge_features
        self.dim_target = dim_target
        self.config = config

    def forward(self,
                node_embeddings: torch.tensor,
                batch: torch.Tensor,
                **kwargs) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[List[object]]]:
        r"""
        Performs a forward pass over a batch of graphs

        Args:
            node_embeddings (:class:`torch_geometric.data.Batch`): the node embeddings
            batch (:class:`torch.Tensor`): the usual ``batch`` object of PyG
            kwargs (dict): additional and optional arguments

        Returns:
            a tuple (model's output, [optional] node embeddings, [optional] additional outputs
        """
        raise NotImplementedError('You need to implement this method!')


import torch
from torch_geometric.utils import to_dense_adj, to_dense_batch

from pydgn.model.interface import ReadoutInterface


class DotProductLinkReadout(ReadoutInterface):
    """
    Class that implements a simple readout mapping for link prediction via dot product
    """
    def forward(self, node_embeddings, batch, **kwargs):
        edge_index = kwargs['edge_index']

        z, _ = to_dense_batch(node_embeddings, batch)
        adj = to_dense_adj(edge_index, batch)

        batch_size, num_nodes, _ = z.size()
        '''
        >>> As = torch.randn(3,2,5)
        >>> Bs = torch.randn(3,5,4)
        >>> torch.einsum('bij,bjk->bik', As, Bs) # batch matrix multiplication
        s = F.sigmoid(torch.einsum('bij,bji->bii', x, torch.transpose(x)))  # batch matrix multiplication
        '''
        return None, node_embeddings, torch.sigmoid(torch.matmul(z, z.transpose(1, 2))), adj



import os

from torch.utils.data import DataLoader
from torch_geometric.data import Batch

from pydgn.experiment.experiment import Experiment
from pydgn.static import LOSS, SCORE


class SemiSupervisedTask(Experiment):
    """
    Class that implements a semi-supervised experiment. There is an ``unsupervised_config`` field in the configuration
    file that is used, together with the field ``model``, to produce unsupervised embeddings. These are later used
    by a **readout** extracted from a ``supervised_config`` field in the configuration file to perform the supervised
    task.
    """
    def __init__(self, model_configuration, exp_path, exp_seed):
        super(SemiSupervisedTask, self).__init__(model_configuration, exp_path, exp_seed)
        self.root_exp_path = exp_path  # to distinguish unsup. and sup. exp paths

    def run_valid(self, dataset_getter, logger):
        unsupervised_config = self.model_config.unsupervised_config
        supervised_config = self.model_config.supervised_config

        # -------------------------------------- PART I: Unsupervised Training --------------------------------------- #

        batch_size = unsupervised_config['batch_size']
        shuffle = unsupervised_config['shuffle'] \
            if 'shuffle' in unsupervised_config else True

        # Instantiate the Dataset
        train_loader = dataset_getter.get_inner_train(batch_size=batch_size, shuffle=shuffle)
        val_loader = dataset_getter.get_inner_val(batch_size=batch_size, shuffle=shuffle)

        # Call this after the loaders: the datasets may need to be instantiated with additional parameters
        dim_node_features = dataset_getter.get_dim_node_features()
        dim_edge_features = dataset_getter.get_dim_edge_features()
        dim_target = dataset_getter.get_dim_target()

        # Change exp path to allow Stop & Resume
        self.exp_path = os.path.join(self.root_exp_path, f'unsupervised_training')

        # Istantiate the Model
        model = self.create_unsupervised_model(dim_node_features, dim_edge_features, dim_target)

        # Istantiate the engine (it handles the training loop and the inference phase by abstracting the specifics)
        unsupervised_training_engine = self.create_unsupervised_engine(model)

        _, _, train_data_list, \
        _, _, val_data_list, \
        _, _, _ = unsupervised_training_engine.train(train_loader=train_loader,
                                                      validation_loader=val_loader,
                                                      test_loader=None,
                                                      max_epochs=unsupervised_config['epochs'],
                                                      logger=logger)

        # --------------------------------------- PART II: Supervised Training --------------------------------------- #

        # Get the embedding size from the first graph in the dataset
        embedding_size = train_data_list[0].x.shape[1]

        # Get information from the supervised configuration
        batch_size = supervised_config['batch_size']
        shuffle = supervised_config['shuffle'] \
            if 'shuffle' in supervised_config else True
        collate_fn = lambda data_list: Batch.from_data_list(data_list)

        # Instantiate the Embedding Dataset for supervised learning
        train_loader = DataLoader(train_data_list, batch_size, shuffle, collate_fn=collate_fn)
        val_loader = DataLoader(val_data_list, batch_size, shuffle, collate_fn=collate_fn)

        # Change exp path to allow Stop & Resume
        self.exp_path = os.path.join(self.root_exp_path, f'supervised_training')

        # Instantiate the Model
        model = self.create_supervised_model(dim_node_features=embedding_size, dim_edge_features=0,
                                             dim_target=dim_target)

        # Instantiate the engine (it handles the training loop and the inference phase by abstracting the specifics)
        supervised_training_engine = self.create_supervised_engine(model)

        train_loss, train_score, _, \
        val_loss, val_score, _, \
        _, _, _ = supervised_training_engine.train(train_loader=train_loader,
                                                    validation_loader=val_loader,
                                                    test_loader=None,
                                                    max_epochs=supervised_config['epochs'],
                                                    logger=logger)

        train_res = {LOSS: train_loss, SCORE: train_score}
        val_res = {LOSS: val_loss, SCORE: val_score}
        return train_res, val_res

    def run_test(self, dataset_getter, logger):
        unsupervised_config = self.model_config.unsupervised_config
        supervised_config = self.model_config.supervised_config

        # -------------------------------------- PART I: Unsupervised Training --------------------------------------- #

        batch_size = unsupervised_config['batch_size']
        shuffle = unsupervised_config['shuffle'] \
            if 'shuffle' in unsupervised_config else True

        # Instantiate the Dataset
        train_loader = dataset_getter.get_outer_train(batch_size=batch_size, shuffle=shuffle)
        val_loader = dataset_getter.get_outer_val(batch_size=batch_size, shuffle=shuffle)
        test_loader = dataset_getter.get_outer_test(batch_size=batch_size, shuffle=shuffle)

        # Call this after the loaders: the datasets may need to be instantiated with additional parameters
        dim_node_features = dataset_getter.get_dim_node_features()
        dim_edge_features = dataset_getter.get_dim_edge_features()
        dim_target = dataset_getter.get_dim_target()

        # Change exp path to allow Stop & Resume
        self.exp_path = os.path.join(self.root_exp_path, f'unsupervised_training')

        # Instantiate the Model
        model = self.create_unsupervised_model(dim_node_features, dim_edge_features, dim_target)

        # Instantiate the engine (it handles the training loop and the inference phase by abstracting the specifics)
        unsupervised_training_engine = self.create_unsupervised_engine(model)

        _, _, train_data_list, \
        _, _, val_data_list, \
        _, _, test_data_list = unsupervised_training_engine.train(train_loader=train_loader,
                                                                   validation_loader=val_loader,
                                                                   test_loader=test_loader,
                                                                   max_epochs=unsupervised_config['epochs'],
                                                                   logger=logger)

        # --------------------------------------- PART II: Supervised Training --------------------------------------- #

        # Get the embedding size from the first graph in the dataset
        embedding_size = train_data_list[0].x.shape[1]

        # Get information from the supervised configuration
        batch_size = supervised_config['batch_size']
        shuffle = supervised_config['shuffle'] \
            if 'shuffle' in supervised_config else True
        collate_fn = lambda data_list: Batch.from_data_list(data_list)

        # Instantiate the Embedding Dataset for supervised learning
        train_loader = DataLoader(train_data_list, batch_size, shuffle, collate_fn=collate_fn)
        val_loader = DataLoader(val_data_list, batch_size, shuffle, collate_fn=collate_fn)
        test_loader = DataLoader(test_data_list, batch_size, shuffle, collate_fn=collate_fn)

        # Change exp path to allow Stop & Resume
        self.exp_path = os.path.join(self.root_exp_path, f'supervised_training')

        # Instantiate the Model
        model = self.create_supervised_model(dim_node_features=embedding_size, dim_edge_features=0,
                                             dim_target=dim_target)

        # Instantiate the engine (it handles the training loop and the inference phase by abstracting the specifics)
        supervised_training_engine = self.create_supervised_engine(model)

        train_loss, train_score, _, \
        val_loss, val_score, _, \
        test_loss, test_score, _ = supervised_training_engine.train(train_loader=train_loader,
                                                                     validation_loader=val_loader,
                                                                     test_loader=test_loader,
                                                                     max_epochs=supervised_config['epochs'],
                                                                     logger=logger)

        train_res = {LOSS: train_loss, SCORE: train_score}
        val_res = {LOSS: val_loss, SCORE: val_score}
        test_res = {LOSS: test_loss, SCORE: test_score}
        return train_res, val_res, test_res



import os
import sys

import gpustat

import yaml
import argparse

from pydgn.static import *


def set_gpus(num_gpus):
    """
    Sets the visible GPUS for the experiments according to the availability in terms of memory. Prioritize GPUs with
    less memory usage. Sets the ``CUDA_DEVICE_ORDER`` env variable to ``PCI_BUS_ID`` and ``CUDA_VISIBLE_DEVICES``
    to the ordered list of GPU indices.

    Args:
        num_gpus: maximum number of GPUs to use when launching experiments in parallel
    """
    try:
        selected = []

        stats = gpustat.GPUStatCollection.new_query()

        for i in range(num_gpus):

            ids_mem = [res for res in map(lambda gpu: (int(gpu.entry['index']),
                                                       float(gpu.entry['memory.used']) / \
                                                       float(gpu.entry['memory.total'])),
                                          stats) if str(res[0]) not in selected]

            if len(ids_mem) == 0:
                # No more gpus available
                break

            best = min(ids_mem, key=lambda x: x[1])
            bestGPU, bestMem = best[0], best[1]
            # print(f"{i}-th best is {bestGPU} with mem {bestMem}")
            selected.append(str(bestGPU))

        print("Setting GPUs to: {}".format(",".join(selected)))
        os.environ['CUDA_DEVICE_ORDER'] = 'PCI_BUS_ID'
        os.environ['CUDA_VISIBLE_DEVICES'] = ",".join(selected)
    except BaseException as e:
        print("GPU not available: " + str(e))


def evaluation(options):
    kwargs = vars(options)
    debug = kwargs[DEBUG]
    configs_dict = yaml.load(open(kwargs[CONFIG_FILE], "r"), Loader=yaml.FullLoader)

    # Hardware initialization
    max_cpus = configs_dict[MAX_CPUS]

    device = configs_dict[DEVICE]
    use_cuda = CUDA in device

    if not use_cuda:
        max_gpus = 0
        gpus_per_task = 0
    else:
        # We will choose the GPU with least ratio of memory usage
        max_gpus = configs_dict[MAX_GPUS]
        # Choose which GPUs to use
        set_gpus(max_gpus)
        gpus_per_task = configs_dict[GPUS_PER_TASK]

    # we probably don't need this anymore, but keep it commented in case we're wrong
    # OMP_NUM_THREADS = 'OMP_NUM_THREADS'
    # os.environ[OMP_NUM_THREADS] = "1"  # This is CRUCIAL to avoid bottlenecks when running experiments in parallel. DO NOT REMOVE IT

    #
    # Once CUDA_VISIBLE_DEVICES has been set, we can start importing all the necessary modules
    #
    import ray
    import torch

    # we probably don't need this anymore, but keep it commented in case we're wrong
    # Needed to avoid thread spawning, conflicts with multi-processing. You may set a number > 1 but take into account the number of processes on the machine
    # torch.set_num_threads(1)

    from pydgn.experiment.util import s2c
    from pydgn.data.splitter import Splitter
    from pydgn.evaluation.grid import Grid
    from pydgn.evaluation.random_search import RandomSearch

    assert GRID_SEARCH in configs_dict or RANDOM_SEARCH in configs_dict
    search_class = (Grid if GRID_SEARCH in configs_dict else RandomSearch)
    search = search_class(configs_dict)

    if use_cuda:
        # Ensure a generic "cuda" device is set when using more than 1 GPU
        search.device = CUDA

    # Set the random seed
    seed = search.seed if search.seed is not None else 42
    print(f'Base seed set to {seed}.')
    experiment = search.experiment
    experiment_class = s2c(experiment)
    exp_path = os.path.join(configs_dict[RESULT_FOLDER], f"{search.exp_name}")

    os.environ[PYDGN_RAY_NUM_GPUS_PER_TASK] = str(float(gpus_per_task))

    # You can make PyDGN work on a cluster of machines!
    if os.environ.get('ip_head') is not None:
        assert os.environ.get('redis_password') is not None
        ray.init(address=os.environ.get('ip_head'), _redis_password=os.environ.get('redis_password'))
        print("Connected to Ray cluster.")
        print(f"Available nodes: {ray.nodes()}")
    # Or you can work on your single server
    else:
        ray.init(num_cpus=max_cpus, num_gpus=max_gpus)
        print(f"Started local ray instance.")

    data_splits_file = configs_dict[DATA_SPLITS_FILE]

    splitter = Splitter.load(data_splits_file)
    inner_folds, outer_folds = splitter.n_inner_folds, splitter.n_outer_folds
    print(f'Data splits loaded, outer folds are {outer_folds} and inner folds are {inner_folds}')

    # WARNING: leave the import here, it reads env variables set before
    from pydgn.evaluation.evaluator import RiskAssesser
    risk_assesser = RiskAssesser(outer_folds,
                                 inner_folds,
                                 experiment_class,
                                 exp_path,
                                 data_splits_file,
                                 search,
                                 final_training_runs=configs_dict[FINAL_TRAINING_RUNS],
                                 higher_is_better=search.higher_results_are_better,
                                 gpus_per_task=gpus_per_task,
                                 base_seed=seed)

    risk_assesser.risk_assessment(debug=debug)


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(CONFIG_FILE_CLI_ARGUMENT, dest=CONFIG_FILE)
    parser.add_argument(DEBUG_CLI_ARGUMENT, action="store_true", dest=DEBUG, default=False)
    return parser.parse_args()


def main():
    # Necessary to locate dotted paths in projects that use PyDGN
    sys.path.append(os.getcwd())

    options = get_args()
    try:
        evaluation(options)
    except Exception as e:
        raise e



import setuptools

__version__ = "1.1.0"

URL = 'https://pydgn.readthedocs.io/en/latest/'

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pydgn",
    version=__version__,
    author="Federico Errica",
    author_email="f.errica@protonmail.com",
    description="A Python Package for Deep Graph Networks",
    long_description_content_type="text/markdown",
    url=URL,
    keywords=[
        'deep-graph-networks',
        'evaluation-framework',
        'deep-learning-for-graphs'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires=">=3.8,<3.10",
    install_requires=[
        'PyYAML >= 5.4',
        'networkx >= 2.3',
        'requests >= 2.22.0',
        'matplotlib >= 3.3.4',
        'seaborn >= 0.9.0',
        'tensorboard >= 2.1.1',
        'tqdm >= 4.47.0',
        'ogb >= 1.3.3',
        'aioredis >= 1.3.1',
        'ray >= 1.5.2',
        'gpustat >= 0.6.0',
        'torch >= 1.10.2',
        'torch-geometric >= 2.0.3',
        'torch-geometric-temporal >= 0.52.0',
    ],
    packages=setuptools.find_packages()
)



from typing import List, Union

import torch
from torch.nn import Module, CrossEntropyLoss, MSELoss, L1Loss

from pydgn.experiment.util import s2c
from pydgn.static import BATCH_LOSS_EXTRA, ARGS, CLASS_NAME
from pydgn.training.event.handler import EventHandler
from pydgn.training.event.state import State


class Metric(Module, EventHandler):
    r"""
    Metric is the main event handler for all metrics. Other metrics can easily subclass by implementing the :func:`forward`
    method, though sometimes more complex implementations are required.

    Args:
        use_as_loss (bool): whether this metric should act as a loss (i.e., it should act when :func:`on_backward` is called). **Used by PyDGN, no need to care about this.**
        reduction (str): the type of reduction to apply across samples of the mini-batch. Supports ``mean`` and ``sum``. Default is ``mean``.
        use_nodes_batch_size (bool): whether or not to use the # of nodes in the batch, rather than the number of graphs, to compute
        the metric's aggregated value for the entire epoch.
        accumulate_over_time_steps (bool): [temporal graph learning] whether or not we should accumulate results across different time steps in a batch.
    """
    def __init__(self, use_as_loss: bool=False, reduction: str='mean',
                 use_nodes_batch_size: bool=False,
                 accumulate_over_time_steps: bool=False):
        super().__init__()
        self.batch_metrics = None
        self.num_samples = None
        self.cumulative_batch_num_samples = None
        self.use_as_loss = use_as_loss
        self.reduction = reduction
        self.use_nodes_batch_size = use_nodes_batch_size
        self.accumulate_over_time_steps = accumulate_over_time_steps

    @property
    def name(self) -> str:
        raise NotImplementedError('You should subclass Metric and implement this method!')

    def get_main_metric_name(self) -> str:
        """
        Return the metric's main name. Useful when a metric is the combination of many.

        Returns:
            the metric's main name
        """
        return self.name

    def _expand_reduction(self, state: State):
        #
        # Returns the number of samples to use, conditioned on the type of reduction (e.g., sum, mean) to obtain
        # the sum of the individual sample scores.
        # Used to average sample scores across the entire epoch, rather than taking an average of minibatch scores
        #
        if self.reduction == 'mean':
            # Used to recover the "sum" version of the metric
            return self._update_num_samples(state)

        elif self.reduction == 'sum':
            return 1
        else:
            raise NotImplementedError('The only reductions allowed are sum and mean')

    def _compute_num_samples(self, state: State) -> int:
        """
        Returns the number of samples to consider in a batch when scaling the loss/score.

        Args:
            state (:class:`~training.event.state.State`): object holding training information

        Returns:
            The number of samples to use when scaling the loss/score.

        """
        return state.batch_num_targets if not self.use_nodes_batch_size else state.batch_num_nodes

    def _update_num_samples(self, state: State) -> int:
        """
        Gives the number of samples to add to the total samples per epoch.
        The method takes into account whether we are performing temporal graph learning or not, by looking at the
        field ``accumulate_over_time_steps``.

        Args:
            state (:class:`~training.event.state.State`): object holding training information

        Returns:
            The number of samples to add to the total samples per epoch.
        """
        if not self.accumulate_over_time_steps:
            return self._compute_num_samples(state)
        else:
            return self.cumulative_batch_num_samples

    def on_training_epoch_start(self, state: State):
        self.batch_metrics = []
        self.num_samples = 0

    def on_training_batch_start(self, state: State):
        self.cumulative_batch_num_samples = 0

    def on_training_batch_end(self, state: State):
        if self.use_as_loss:
            batch_metric = state.batch_loss
        else:
            batch_metric = state.batch_score

        self.batch_metrics.append(batch_metric[self.name].item() * self._expand_reduction(state))
        self.num_samples += self._update_num_samples(state)
        self.cumulative_batch_num_samples = None

    def on_training_epoch_end(self, state: State):
        if self.use_as_loss:
            state.update(epoch_loss={self.name: torch.tensor(self.batch_metrics).sum() / self.num_samples})
        else:
            state.update(epoch_score={self.name: torch.tensor(self.batch_metrics).sum() / self.num_samples})

        self.batch_metrics = None
        self.num_samples = None

    def on_eval_batch_start(self, state: State):
        self.cumulative_batch_num_samples = 0

    def on_eval_epoch_start(self, state: State):
        self.batch_metrics = []
        self.num_samples = 0

    def on_eval_epoch_end(self, state: State):
        if self.use_as_loss:
            state.update(epoch_loss={self.name: torch.tensor(self.batch_metrics).sum() / self.num_samples})
        else:
            state.update(epoch_score={self.name: torch.tensor(self.batch_metrics).sum() / self.num_samples})

        self.batch_metrics = None
        self.num_samples = None

    def on_eval_batch_end(self, state: State):
        if self.use_as_loss:
            batch_metric = state.batch_loss
        else:
            batch_metric = state.batch_score

        self.batch_metrics.append(batch_metric[self.name].item() * self._expand_reduction(state))
        self.num_samples += self._update_num_samples(state)
        self.cumulative_batch_num_samples = None

    def on_compute_metrics(self, state: State):
        outputs, targets = state.batch_outputs, state.batch_targets

        if self.use_as_loss:
            loss_output = self.forward(targets, *outputs)

            if isinstance(loss_output, tuple):
                # Allow loss to produce intermediate results that speed up
                # Score computation. This loss callback MUST occur before the score one.
                loss, extra = loss_output
                state.update(batch_loss_extra={self.name: extra})
            else:
                loss = loss_output

            # [temporal graph learning] accumulate, rather than substitute, the losses across multiple snapshots
            if self.accumulate_over_time_steps and state.batch_loss is not None:
                old_loss = state.batch_loss
                loss = old_loss[self.name] + loss

            state.update(batch_loss={self.name: loss})

        else:
            score = self.forward(targets,
                                 *outputs,
                                 batch_loss_extra=getattr(state, BATCH_LOSS_EXTRA, None))

            # Handle base case: forward returns a score
            # This is used to work with implementations that return multiple scores, such as MultiScore
            if type(score) is not dict:
                score = {self.name: score}

            # Score is a dictionary with key-value pairs
            # we need to detach each score from the computational graph
            score = {k: v.detach().cpu() for k, v in score.items()}

            # [temporal graph learning] accumulate, rather than substitute, the scores across multiple snapshots
            if self.accumulate_over_time_steps and state.batch_score is not None:
                old_score = state.batch_score
                score = {k: old_score[k] + score[k] for k in score.keys()}

            state.update(batch_score=score)

        # [temporal graph learning] accumulate the number of samples of each snapshot for each batch
        if self.accumulate_over_time_steps:
            self.cumulative_batch_num_samples += self._compute_num_samples(state)

    def on_backward(self, state: State):
        if self.use_as_loss:
            try:
                state.batch_loss[self.name].backward()
            except Exception as e:
                # Here we catch potential multiprocessing related issues
                # see https://github.com/pytorch/pytorch/wiki/Autograd-and-Fork
                    print(e)

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict=None) -> Union[dict, float]:
        r"""
        Computes the metric value. Optionally, and only for scores used as losses, some extra information can be also returned.

        Args:
            targets (:class:`torch.Tensor`): ground truth
            outputs (List[:class:`torch.Tensor`]): outputs of the model
            batch_loss_extra (dict): dictionary of information computed by metrics used as losses

        Returns:
            A dictionary containing associations metric_name - value or simply a value
        """
        raise NotImplementedError('To be implemented by a subclass')


class MultiScore(Metric):
    r"""
    This class is used to keep track of multiple additional metrics used as scores, rather than losses.

    Args:
        use_as_loss (bool): whether this metric should act as a loss (i.e., it should act when :func:`on_backward` is called). **Used by PyDGN, no need to care about this.**
        reduction (str): the type of reduction to apply across samples of the mini-batch. Supports ``mean`` and ``sum``. Default is ``mean``.
        use_nodes_batch_size (bool): whether or not to use the # of nodes in the batch, rather than the number of graphs, to compute
        the metric's aggregated value for the entire epoch.
        accumulate_over_time_steps (bool): [temporal graph learning] whether or not we should accumulate results across different time steps.

    """
    def __init__(self, use_as_loss, reduction='mean',
                 use_nodes_batch_size=False, accumulate_over_time_steps: bool=False,
                 main_scorer=None, **extra_scorers):

        assert not use_as_loss, "MultiScore cannot be used as loss"
        assert not main_scorer is None, "You have to provide a main scorer"
        super().__init__(use_as_loss, reduction, use_nodes_batch_size, accumulate_over_time_steps)
        self.scorers = [self._istantiate_scorer(main_scorer)] + [self._istantiate_scorer(score) for score in
                                                                 extra_scorers.values()]

    @property
    def name(self) -> str:
        return "Multi Score"

    def _istantiate_scorer(self, scorer):
        if isinstance(scorer, dict):
            args = scorer[ARGS]
            return s2c(scorer[CLASS_NAME])(use_as_loss=False, **args)
        else:
            return s2c(scorer)(use_as_loss=False)

    def get_main_metric_name(self):
        return self.scorers[0].get_main_metric_name()

    def on_training_epoch_start(self, state: State):
        self.batch_metrics = {s.name: [] for s in self.scorers}
        for scorer in self.scorers:
            scorer.on_training_epoch_start(state)

    def on_training_batch_end(self, state: State):
        for scorer in self.scorers:
            scorer.on_training_batch_end(state)

    def on_training_epoch_end(self, state: State):
        epoch_score = {}
        for scorer in self.scorers:
            # This will update the epoch_score variable in State
            scorer.on_training_epoch_end(state)
            epoch_score.update(state.epoch_score)
        state.update(epoch_score=epoch_score)

    def on_eval_epoch_start(self, state: State):
        for scorer in self.scorers:
            scorer.on_eval_epoch_start(state)

    def on_eval_batch_end(self, state: State):
        for scorer in self.scorers:
            scorer.on_eval_batch_end(state)

    def on_eval_epoch_end(self, state: State):
        epoch_score = {}
        for scorer in self.scorers:
            # This will update the epoch_score variable in State
            scorer.on_training_epoch_end(state)
            epoch_score.update(state.epoch_score)
        state.update(epoch_score=epoch_score)

    def on_compute_metrics(self, state: State):
        super().on_compute_metrics(state)

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict=None) -> dict:
        res = {}
        for scorer in self.scorers:
            # each scorer __call__ method returns a dict
            res.update({scorer.name: scorer.forward(targets, *outputs, batch_loss_extra=batch_loss_extra)})

        return res


class AdditiveLoss(Metric):
    r"""
    AdditiveLoss sums an arbitrary number of losses together.

    Args:
        use_as_loss (bool): whether this metric should act as a loss (i.e., it should act when :func:`on_backward` is called). **Used by PyDGN, no need to care about this.**
        reduction (str): the type of reduction to apply across samples of the mini-batch. Supports ``mean`` and ``sum``. Default is ``mean``.
        use_nodes_batch_size (bool): whether or not to use the # of nodes in the batch, rather than the number of graphs, to compute
        accumulate_over_time_steps (bool): [temporal graph learning] whether or not we should accumulate results across different time steps.
        the metric's aggregated value for the entire epoch.
        losses (dict): dictionary of metrics to add together
    """
    def __init__(self, use_as_loss, reduction='mean',
                 use_nodes_batch_size=False, accumulate_over_time_steps: bool=False, **losses: dict):
        assert use_as_loss, "Additive loss can only be used as a loss"
        super().__init__(use_as_loss, reduction, use_nodes_batch_size, accumulate_over_time_steps)

        self.losses = [self._instantiate_loss(loss) for loss in losses.values()]
        self.use_nodes_batch_size = use_nodes_batch_size
        self.accumulate_over_time_steps = accumulate_over_time_steps

    @property
    def name(self) -> str:
        return "Additive Loss"

    def _instantiate_loss(self, loss):
        if isinstance(loss, dict):
            args = loss[ARGS]
            return s2c(loss[CLASS_NAME])(use_as_loss=True, **args)
        else:
            return s2c(loss)(use_as_loss=True)

    def on_training_epoch_start(self, state: State):
        self.batch_metrics = {l.name: [] for l in [self] + self.losses}
        self.num_samples = 0
        self.cumulatime_time_steps_samples = 0

    def on_training_batch_end(self, state: State):
        for k, v in state.batch_loss.items():
            self.batch_metrics[k].append(v.item() * self._expand_reduction(state))
        self.num_samples += self._update_num_samples(state)
        self.cumulative_batch_num_samples = None


    def on_training_epoch_end(self, state: State):
        state.update(epoch_loss={l.name: torch.tensor(self.batch_metrics[l.name]).sum() / self.num_samples
                                 for l in [self] + self.losses})
        self.batch_metrics = None
        self.num_samples = None

    def on_eval_epoch_start(self, state: State):
        self.batch_metrics = {l.name: [] for l in [self] + self.losses}
        self.num_samples = 0
        self.cumulatime_time_steps_samples = 0

    def on_eval_epoch_end(self, state: State):
        state.update(epoch_loss={l.name: torch.tensor(self.batch_metrics[l.name]).sum() / self.num_samples
                                 for l in [self] + self.losses})

        self.batch_metrics = None
        self.num_samples = None

    def on_eval_batch_end(self, state: State):
        for k, v in state.batch_loss.items():
            self.batch_metrics[k].append(v.item() * self._expand_reduction(state))
        self.num_samples += self._update_num_samples(state)
        self.cumulative_batch_num_samples = None


    def on_compute_metrics(self, state: State):
        outputs, targets = state.batch_outputs, state.batch_targets
        loss = {}
        extra = {}
        loss_sum = 0.
        for l in self.losses:
            single_loss = l.forward(targets, *outputs)
            if isinstance(single_loss, tuple):
                # Allow loss to produce intermediate results that speed up
                # Score computation. Loss callback MUST occur before the score one.
                loss_output, loss_extra = single_loss
                extra[l.name] = loss_extra
                state.update(batch_loss_extra=extra)
            else:
                loss_output = single_loss

            loss[l.name] = loss_output
            loss_sum += loss_output

        loss[self.name] = loss_sum

        # [temporal graph learning] accumulate, rather than substitute, the losses across multiple snapshots
        if self.accumulate_over_time_steps and state.batch_loss is not None:
            old_loss = state.batch_loss
            loss = {k: old_loss[k] + loss[k] for k in loss.keys()}

            # [temporal graph learning] accumulate the number of samples of each snapshot for each batch
            self.cumulative_batch_num_samples += self._compute_num_samples(state)

        state.update(batch_loss=loss)

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict=None) -> dict:
        pass
    

class Classification(Metric):
    r"""
    Generic metric for classification tasks. Used to maximize code reuse for classical metrics.
    """
    def __init__(self, use_as_loss=False, reduction='mean',
                 use_nodes_batch_size=False, accumulate_over_time_steps: bool=False):
        super().__init__(use_as_loss=use_as_loss, reduction=reduction,
                         use_nodes_batch_size=use_nodes_batch_size, accumulate_over_time_steps=accumulate_over_time_steps)
        self.metric = None

    @property
    def name(self) -> str:
        return 'Classification Metric'

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict=None) -> dict:
        outputs = outputs[0]

        if len(targets.shape) == 2:
            targets = targets.squeeze(dim=1)

        metric = self.metric(outputs.squeeze(dim=1), targets)
        return metric


class Regression(Metric):
    r"""
    Generic metric for regression tasks. Used to maximize code reuse for classical metrics.
    """
    def __init__(self, use_as_loss=False, reduction='mean',
                 use_nodes_batch_size=False, accumulate_over_time_steps: bool=False):
        super().__init__(use_as_loss=use_as_loss, reduction=reduction,
                         use_nodes_batch_size=use_nodes_batch_size, accumulate_over_time_steps=accumulate_over_time_steps)
        self.metric = None

    @property
    def name(self) -> str:
        return 'Regression Metric'

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict=None) -> dict:
        outputs = outputs[0]

        if len(targets.shape) == 2:
            targets = targets.squeeze(dim=1)

        metric = self.metric(outputs.squeeze(dim=1), targets)
        return metric


class MulticlassClassification(Classification):
    r"""
    Wrapper around :class:`torch.nn.CrossEntropyLoss`
    """
    def __init__(self, use_as_loss=False, reduction='mean',
                 use_nodes_batch_size=False, accumulate_over_time_steps: bool=False):
        super().__init__(use_as_loss=use_as_loss, reduction=reduction,
                         use_nodes_batch_size=use_nodes_batch_size, accumulate_over_time_steps=accumulate_over_time_steps)
        self.metric = CrossEntropyLoss(reduction=reduction)

    @property
    def name(self) -> str:
        return 'Multiclass Classification'


class MeanSquareError(Regression):
    r"""
    Wrapper around :class:`torch.nn.MSELoss`
    """
    def __init__(self, use_as_loss=False, reduction='mean',
                 use_nodes_batch_size=False, accumulate_over_time_steps: bool=False):
        super().__init__(use_as_loss=use_as_loss, reduction=reduction,
                         use_nodes_batch_size=use_nodes_batch_size, accumulate_over_time_steps=accumulate_over_time_steps)
        self.metric = MSELoss(reduction=reduction)

    @property
    def name(self) -> str:
        return 'Mean Square Error'


class MeanAverageError(Regression):
    r"""
    Wrapper around :class:`torch.nn.MSELoss`
    """
    def __init__(self, use_as_loss=False, reduction='mean',
                 use_nodes_batch_size=False, accumulate_over_time_steps: bool=False):
        super().__init__(use_as_loss=use_as_loss, reduction=reduction,
                         use_nodes_batch_size=use_nodes_batch_size, accumulate_over_time_steps=accumulate_over_time_steps)
        self.metric = L1Loss(reduction=reduction)

    @property
    def name(self) -> str:
        return 'Mean Average Error'


class DotProductLink(Metric):
    """
    Implements a dot product link prediction metric, as defined in https://arxiv.org/abs/1611.07308.
    """
    @property
    def name(self) -> str:
        return 'Dot Product Link Prediction'

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict=None) -> dict:
        node_embs = outputs[1]
        _, pos_edges, neg_edges = targets

        loss_edge_index = torch.cat((pos_edges, neg_edges), dim=1)
        loss_target = torch.cat((torch.ones(pos_edges.shape[1]),
                                 torch.zeros(neg_edges.shape[1])))

        # Taken from https://github.com/rusty1s/pytorch_geometric/blob/master/examples/link_pred.py
        x_j = torch.index_select(node_embs, 0, loss_edge_index[0])
        x_i = torch.index_select(node_embs, 0, loss_edge_index[1])
        link_logits = torch.einsum("ef,ef->e", x_i, x_j)
        loss = torch.nn.functional.binary_cross_entropy_with_logits(link_logits, loss_target.to(link_logits.device))
        return loss


class MulticlassAccuracy(Metric):
    """
    Implements multiclass classification accuracy.
    """

    @property
    def name(self) -> str:
        return 'Multiclass Accuracy'

    @staticmethod
    def _get_correct(output):
        return torch.argmax(output, dim=1)

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict=None) -> dict:
        pred = outputs[0]
        correct = self._get_correct(pred)

        if len(targets.shape) == 2:
            targets = targets.squeeze(dim=1)

        return 100. * (correct == targets).sum().float() / targets.size(0)


class ToyMetric(Metric):
    r"""
    Implements a toy metric.
    """

    def __init__(self, use_as_loss=False, reduction='mean',
                 use_nodes_batch_size=False, accumulate_over_time_steps: bool=False):
        super().__init__(use_as_loss=use_as_loss, reduction=reduction,
                         use_nodes_batch_size=use_nodes_batch_size, accumulate_over_time_steps=accumulate_over_time_steps)

    @property
    def name(self) -> str:
        return 'Toy Metric'

    @staticmethod
    def _get_correct(output):
        return torch.argmax(output, dim=1)

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict=None) -> dict:
        return 100. * torch.ones(1)


class ToyUnsupervisedMetric(Metric):

    @property
    def name(self) -> str:
        return 'Toy Unsupervised Metric'

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict = None) -> dict:
        return (outputs[0]*0.).mean()


from pydgn.experiment.experiment import Experiment
from pydgn.static import LOSS, SCORE


class SupervisedTask(Experiment):
    """
    Class that implements a standard supervised experiment.
    """
    def run_valid(self, dataset_getter, logger):
        batch_size = self.model_config.supervised_config['batch_size']
        shuffle = self.model_config.supervised_config['shuffle'] \
            if 'shuffle' in self.model_config.supervised_config else True

        # Instantiate the Dataset
        train_loader = dataset_getter.get_inner_train(batch_size=batch_size, shuffle=shuffle)
        val_loader = dataset_getter.get_inner_val(batch_size=batch_size, shuffle=shuffle)

        dim_node_features = dataset_getter.get_dim_node_features()
        dim_edge_features = dataset_getter.get_dim_edge_features()
        dim_target = dataset_getter.get_dim_target()

        # Instantiate the Model
        model = self.create_supervised_model(dim_node_features, dim_edge_features, dim_target)

        # Instantiate the engine (it handles the training loop and the inference phase by abstracting the specifics)
        supervised_training_engine = self.create_supervised_engine(model)

        train_loss, train_score, _, \
        val_loss, val_score, _, \
        _, _, _ = supervised_training_engine.train(
            train_loader=train_loader,
            validation_loader=val_loader,
            test_loader=None,
            max_epochs=self.model_config.supervised_config['epochs'],
            logger=logger)

        train_res = {LOSS: train_loss, SCORE: train_score}
        val_res = {LOSS: val_loss, SCORE: val_score}
        return train_res, val_res

    def run_test(self, dataset_getter, logger):
        batch_size = self.model_config.supervised_config['batch_size']
        shuffle = self.model_config.supervised_config['shuffle'] \
            if 'shuffle' in self.model_config.supervised_config else True

        # Instantiate the Dataset
        train_loader = dataset_getter.get_outer_train(batch_size=batch_size, shuffle=shuffle)
        val_loader = dataset_getter.get_outer_val(batch_size=batch_size, shuffle=shuffle)
        test_loader = dataset_getter.get_outer_test(batch_size=batch_size, shuffle=shuffle)

        # Call this after the loaders: the datasets may need to be instantiated with additional parameters
        dim_node_features = dataset_getter.get_dim_node_features()
        dim_edge_features = dataset_getter.get_dim_edge_features()
        dim_target = dataset_getter.get_dim_target()

        # Instantiate the Model
        model = self.create_supervised_model(dim_node_features, dim_edge_features, dim_target)

        # Instantiate the engine (it handles the training loop and the inference phase by abstracting the specifics)
        supervised_training_engine = self.create_supervised_engine(model)

        train_loss, train_score, _, \
        val_loss, val_score, _, \
        test_loss, test_score, _ = supervised_training_engine.train(
            train_loader=train_loader,
            validation_loader=val_loader,
            test_loader=test_loader,
            max_epochs=self.model_config.supervised_config['epochs'],
            logger=logger)

        train_res = {LOSS: train_loss, SCORE: train_score}
        val_res = {LOSS: val_loss, SCORE: val_score}
        test_res = {LOSS: test_loss, SCORE: test_score}
        return train_res, val_res, test_res
